"""Unit tests for TokenGeneratorRegistry.

Version Added:
    3.0
"""

from django.test.utils import override_settings

from djblets.secrets.token_generators import token_generator_registry
from djblets.secrets.token_generators.legacy_sha1 import \
    LegacySHA1TokenGenerator
from djblets.secrets.token_generators.vendor_checksum import \
    VendorChecksumTokenGenerator
from djblets.testing.testcases import TestCase


class TokenGeneratorRegistryTests(TestCase):
    """Unit tests for TokenGeneratorRegistry.

    Version Added:
        3.0
    """

    def test_get_default(self):
        """Testing TokenGeneratorRegistry.get_default with no default
        token generator specified in settings
        """
        token_generator = token_generator_registry.get_default()

        self.assertIsInstance(token_generator, VendorChecksumTokenGenerator)

    @override_settings(
        DJBLETS_DEFAULT_API_TOKEN_GENERATOR='legacy_sha1')
    def test_get_default_when_set(self):
        """Testing TokenGeneratorRegistry.get_default with a default
        token generator specified in settings
        """
        token_generator = token_generator_registry.get_default()

        self.assertIsInstance(token_generator, LegacySHA1TokenGenerator)

    @override_settings(
        DJBLETS_DEFAULT_API_TOKEN_GENERATOR='not_registered')
    def test_get_default_when_set_not_registered(self):
        """Testing TokenGeneratorRegistry.get_default with a default token
        generator specified in settings that does not exist in the registry
        """
        expected_log = ('Could not find the not_registered token generator '
                        'in the registry. Using the vendor_checksum token '
                        'generator instead.')

        with self.assertLogs() as logs:
            token_generator = token_generator_registry.get_default()

            self.assertEqual(logs.records[0].getMessage(),
                             expected_log)
            self.assertIsInstance(token_generator,
                                  VendorChecksumTokenGenerator)

    def test_get_deprecated_token_generators(self):
        """Testing TokenGeneratorRegistry.get_deprecated_token_generators with
        no set of deprecated generators specified in settings
        """
        generators = token_generator_registry.get_deprecated_token_generators()

        self.assertSetEqual(generators,
                            {LegacySHA1TokenGenerator.token_generator_id})

    @override_settings(
        DJBLETS_DEPRECATED_TOKEN_GENERATORS={'vendor_checksum'})
    def test_get_deprecated_token_generators_when_set(self):
        """Testing TokenGeneratorRegistry.get_deprecated_token_generators with
        a set of deprecated generators specified in settings
        """
        generators = token_generator_registry.get_deprecated_token_generators()

        self.assertSetEqual(generators,
                            {VendorChecksumTokenGenerator.token_generator_id})



"""Settings API for extensions."""

from django.utils.translation import gettext as _

from djblets.extensions.signals import settings_saved


class ExtensionSettings(dict):
    """Settings data for an extension.

    This is a glorified dictionary that acts as a proxy for the extension's
    stored settings in the database.

    Callers must call :py:meth:`save` when they want to make the settings
    persistent.

    If a key is not found in the dictionary,
    :py:attr:`Extension.default_settings
    <djblets.extensions.extension.Extension.default_settings>` will be checked
    as well.
    """

    def __init__(self, extension):
        """Initialize and load the settings.

        Args:
            extension (djblets.extensions.extension.Extension):
                The extension the settings is for.
        """
        super(ExtensionSettings, self).__init__()

        self.extension = extension
        self.load()

    def __getitem__(self, key):
        """Retrieve an item from the dictionary.

        This will attempt to return a default value from
        :py:attr:`Extension.default_settings
        <djblets.extensions.extension.Extension.default_settings>` if the
        setting has not been set.

        Args:
            key (unicode):
                The key to retrieve.

        Returns:
            object:
            The value from settings.

        Raises:
            KeyError:
                The key could not be found in stored or default settings.
        """
        if super(Settings, self).__contains__(key):
            return super(Settings, self).__getitem__(key)

        if key in self.extension.default_settings:
            return self.extension.default_settings[key]

        raise KeyError(
            _('The settings key "%(key)s" was not found in extension %(ext)s')
            % {
                'key': key,
                'ext': self.extension.id
            })

    def __contains__(self, key):
        """Return if a setting can be found.

        This will check both the stored settings and the default settings
        for the extension.

        Args:
            key (unicode):
                The key to check.

        Returns:
            bool:
            ``True`` if the setting could be found. ``False`` if it could not.
        """
        return (super(Settings, self).__contains__(key) or
                key in self.extension.default_settings)

    def get(self, key, default=None):
        """Return the value for a setting.

        This will return a value from either the stored settings, the
        extensin's default settings, or the value passed as ``default``.

        Args:
            key (unicode):
                The key to retrieve.

            default (object, optional):
                The default value, if it couldn't be found in the stored
                settings or the extension's default settings.

        Returns:
            object:
            The value for the setting.
        """
        # dict.get doesn't call __getitem__ internally, and instead looks up
        # straight from the internal dictionary data. So, we need to handle it
        # ourselves in order to support defaults through __getitem__.
        try:
            return self[key]
        except KeyError:
            return default

    def set(self, key, value):
        """Set a setting's value.

        This is equivalent to setting the value through standard dictionary
        attribute storage.

        Args:
            key (unicode):
                The key to set.

            value (object):
                The value for the setting.
        """
        self[key] = value

    def load(self):
        """Load the settings from the database."""
        try:
            self.update(self.extension.registration.settings)
        except ValueError:
            # The settings in the database are invalid. We'll have to discard
            # it. Note that this should never happen unless the user
            # hand-modifies the entries and breaks something.
            pass

    def save(self):
        """Save all current settings to the database."""
        registration = self.extension.registration
        registration.settings = dict(self)
        registration.save()

        settings_saved.send(sender=self.extension)

        # Make sure others are aware that the configuration changed.
        self.extension.extension_manager._bump_sync_gen()


#: Legacy name for ExtensionSettings.
#:
#: This is unlikely to be needed outside of the Djblets Extensions code, but
#: is available for any callers who might be dependent on it.
#:
#: Deprecated:
#:     2.0:
#:     Renamed to :py:class:`ExtensionSettings`. This will be removed in
#:     Djblets 3.0.
Settings = ExtensionSettings



"""E-mail message composition and sending."""

import logging
from email.utils import parseaddr

from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.utils.datastructures import MultiValueDict
from django.utils.encoding import force_str

from djblets.mail.dmarc import is_email_allowed_by_dmarc
from djblets.mail.utils import build_email_address_via_service


logger = logging.getLogger(__name__)


class EmailMessage(EmailMultiAlternatives):
    """An EmailMesssage subclass with improved header and message ID support.

    This class knows about several headers (standard and variations), including
    :mailheader:`Sender`/:mailheader:`X-Sender`,
    :mailheader:`In-Reply-To`/:mailheader:`References``, and
    :mailheader:`Reply-To`.

    The generated :mailheader:`Message-ID` header from the e-mail can be
    accessed via the :py:attr:`message_id` attribute after the e-mail has been
    sent.

    In order to prevent issues when sending on behalf of users whose e-mail
    domains are controlled by DMARC, callers can specify
    ``from_spoofing`` (or set ``settings.DJBLETS_EMAIL_FROM_SPOOFING``). When
    set, the e-mail address used for the :mailheader:`From` header will only be
    used if there aren't any DMARC rules that may prevent the e-mail from being
    sent/received.

    .. note::

       Releases prior to Djblets 1.0.10 required using
       ``enable_smart_spoofing`` or ``settings.EMAIL_ENABLE_SMART_SPOOFING``,
       which didn't allow From spoofing to be completely disabled.)

    In the event that a DMARC rule would prevent sending on behalf of that
    user, the ``sender`` address will be used instead, with the full name
    appearing as the value in ``from_email`` with "via <Service Name>" tacked
    onto it.

    Callers wishing to use this should also set
    ``settings.EMAIL_DEFAULT_SENDER_SERVICE_NAME`` to the desired service name.
    Otherwise, the domain on the sender e-mail will be used instead.

    This class also supports repeated headers.

    Version Changed:
        1.0.10:
        Added the ``from_spoofing`` parameter and
        ``settings.DJBLETS_EMAIL_FROM_SPOOFING`` to replace
        ``enable_smart_spoofing`` and ``settings.EMAIL_ENABLE_SMART_SPOOFING``.
    """

    #: Always spoof the From address for a user.
    FROM_SPOOFING_ALWAYS = 'always'

    #: Only spoof the From address for a user if allowed by DMARC rules.
    FROM_SPOOFING_SMART = 'smart'

    #: Never spoof the From address for a user.
    FROM_SPOOFING_NEVER = 'never'

    def __init__(self, subject='', text_body='', html_body='', from_email=None,
                 to=None, cc=None, bcc=None, sender=None, in_reply_to=None,
                 headers=None, auto_generated=False,
                 prevent_auto_responses=False, from_spoofing=None,
                 enable_smart_spoofing=None):
        """Create a new EmailMessage.

        Args:
            subject (unicode, optional):
                The subject of the message. Defaults to being blank (which
                MTAs might replace with "no subject".)

            text_body (unicode, optional):
                The body of the e-mail as plain text. Defaults to an empty
                string (allowing HTML-only e-mails to be sent).

            html_body (unicode, optional):
                The body of the e-mail as HTML. Defaults to an empty string
                (allowing text-only e-mails to be sent).

            from_email (unicode, optional):
                The from address for the e-mail. Defaults to
                :django:setting:`DEFAULT_FROM_EMAIL`.

            to (list, optional):
                A list of e-mail addresses as :py:class:`unicode` objects that
                are to receive the e-mail. Defaults to an empty list of
                addresses (allowing using CC/BCC only).

            cc (list, optional):
                A list of e-mail addresses as :py:class:`unicode` objects that
                are to receive a carbon copy of the e-mail, or ``None`` if
                there are no CC recipients.

            bcc (list, optional):
                A list of e-mail addresses as :py:class:`unicode` objects that
                are to receive a blind carbon copy of the e-mail, or ``None``
                if there are not BCC recipients.

            sender (unicode, optional):
                The actual e-mail address sending this e-mail, for use in
                the :mailheader:`Sender` header. If this differs from
                ``from_email``, it will be left out of the header as per
                :rfc:`2822`.

                This will default to :django:setting:`DEFAULT_FROM_EMAIL`
                if unspecified.

            in_reply_to (unicode, optional):
                An optional message ID (which will be used as the value for the
                :mailheader:`In-Reply-To` and :mailheader:`References`
                headers). This will be generated if not provided and will be
                available as the :py:attr:`message_id` attribute after the
                e-mail has been sent.

            headers (django.utils.datastructures.MultiValueDict, optional):
                Extra headers to provide with the e-mail.

            auto_generated (bool, optional):
                If ``True``, the e-mail will contain headers that mark it as
                an auto-generated message (as per :rfc:`3834`) to avoid auto
                replies.

            prevent_auto_responses (bool, optional):
                If ``True``, the e-mail will contain headers to prevent auto
                replies for delivery reports, read receipts, out of office
                e-mails, and other auto-generated e-mails from Exchange.

            from_spoofing (int, optional):
                Whether to spoof the :mailheader:`From` header for the user.
                This can be one of :py:attr:`FROM_SPOOFING_ALWAYS`,
                :py:attr:`FROM_SPOOFING_SMART`, or
                :py:attr:`FROM_SPOOFING_NEVER`.

                This defaults to ``None``, in which case the
                ``enable_smart_spoofing`` will be checked (for legacy reasons),
                falling back to ``settings.DJBLETS_EMAIL_FROM_SPOOFING`` (which
                defaults to :py:attr:`FROM_SPOOFING_ALWAYS`, also for legacy
                reasons).

            enable_smart_spoofing (bool, optional):
                Whether to enable smart spoofing of any e-mail addresses for
                the :mailheader:`From` header (if ``from_spoofing`` is
                ``None``). This defaults to
                ``settings.EMAIL_ENABLE_SMART_SPOOFING``.

                This is deprecated in favor of ``from_spoofing``.
        """
        headers = headers or MultiValueDict()

        if (isinstance(headers, dict) and
            not isinstance(headers, MultiValueDict)):
            # Instantiating a MultiValueDict from a dict does not ensure that
            # values are lists, so we have to ensure that ourselves.
            headers = MultiValueDict(dict(
                (key, [value])
                for key, value in headers.items()
            ))

        if in_reply_to:
            headers['In-Reply-To'] = in_reply_to
            headers['References'] = in_reply_to

        headers['Reply-To'] = from_email

        if from_spoofing is None:
            if enable_smart_spoofing is None:
                enable_smart_spoofing = \
                    getattr(settings, 'EMAIL_ENABLE_SMART_SPOOFING', None)

            if enable_smart_spoofing is not None:
                if enable_smart_spoofing:
                    from_spoofing = self.FROM_SPOOFING_SMART
                else:
                    # This was the original behavior when the setting was
                    # False.
                    from_spoofing = self.FROM_SPOOFING_ALWAYS

            if from_spoofing is None:
                from_spoofing = getattr(settings,
                                        'DJBLETS_EMAIL_FROM_SPOOFING',
                                        self.FROM_SPOOFING_ALWAYS)

        # Figure out the From/Sender we'll be wanting to use.
        if not sender:
            sender = settings.DEFAULT_FROM_EMAIL

        if sender == from_email:
            # RFC 2822 section 3.6.2 states that we should only include Sender
            # if the two are not equal. We also know that we're not spoofing,
            # so e-mail sending should work fine here.
            sender = None
        elif from_spoofing != self.FROM_SPOOFING_ALWAYS:
            # If from_spoofing is in smart mode, we will be checking the
            # DMARC record from the e-mail address we'd be ideally sending on
            # behalf of. If the record indicates that the message has any
            # likelihood of being quarantined or rejected, we'll alter the From
            # field to send using our Sender address instead.
            #
            # If from_spoofing is disabled, we will be changing the from_email
            # to avoid spoofing.
            parsed_from_name, parsed_from_email = parseaddr(from_email)
            parsed_sender_name, parsed_sender_email = parseaddr(sender)

            # The above will return ('', '') if the address couldn't be parsed,
            # so check for this.
            if not parsed_from_email:
                logger.warning('EmailMessage: Unable to parse From address '
                               '"%s"',
                               from_email)

            if not parsed_sender_email:
                logger.warning('EmailMessage: Unable to parse Sender address '
                               '"%s"',
                               sender)

            # We may not be allowed to send on behalf of this user.
            # We actually aren't going to check for this (it may be due
            # to SPF, which is too complex for us to want to check, or
            # it may be due to another ruleset somewhere). Instead, just
            # check if this e-mail could get lost due to the DMARC rules
            # or if from_spoofing is disabled.
            if (from_spoofing == self.FROM_SPOOFING_NEVER or
                (parsed_from_email != parsed_sender_email and
                 not is_email_allowed_by_dmarc(parsed_from_email))):
                # Spoofing is disabled or we can't spoof the e-mail address,
                # so instead, we'll keep the e-mail in Reply To and create a
                # From address we own, which will also indicate what service
                # is sending on behalf of the user.
                from_email = build_email_address_via_service(
                    full_name=parsed_from_name,
                    email=parsed_from_email,
                    sender_email=parsed_sender_email)

        if sender:
            headers['Sender'] = sender
            headers['X-Sender'] = sender

        if auto_generated:
            headers['Auto-Submitted'] = 'auto-generated'

        if prevent_auto_responses:
            headers['X-Auto-Response-Suppress'] = 'DR, RN, OOF, AutoReply'

        # We're always going to explicitly send with the DEFAULT_FROM_EMAIL,
        # but replace the From header with the e-mail address we decided on.
        # While this class and its parent classes don't really care about the
        # difference between these, Django's SMTP e-mail sending machinery
        # treats them differently, sending the value of EmailMessage.from_email
        # when communicating with the SMTP server.
        super(EmailMessage, self).__init__(
            subject=subject,
            body=force_str(text_body),
            from_email=settings.DEFAULT_FROM_EMAIL,
            to=to,
            cc=cc,
            bcc=bcc,
            headers={
                'From': from_email,
            })

        self.message_id = None

        # We don't want to use the regular extra_headers attribute because
        # it will be treated as a plain dict by Django. Instead, since we're
        # using a MultiValueDict, we store it in a separate attribute
        # attribute and handle adding our headers in the message method.
        self._headers = headers

        if html_body:
            self.attach_alternative(force_str(html_body), 'text/html')

    def message(self):
        """Construct an outgoing message for the e-mail.

        This will construct a message based on the data provided to the
        constructor. This represents the e-mail that will later be sent using
        :py:meth:`send`.

        After calling this method, the message's ID will be stored in the
        :py:attr:`message_id` attribute for later reference.

        This does not need to be called manually. It's called by
        :py:meth:`send`.

        Returns:
            django.core.mail.message.SafeMIMEText:
            The resulting message.
        """
        msg = super(EmailMessage, self).message()
        self.message_id = msg['Message-ID']

        for name, value_list in self._headers.lists():
            for value in value_list:
                # Use the native string on each version of Python. These
                # are headers, so they'll be convertible without encoding
                # issues.
                msg[name] = value

        return msg

    def recipients(self):
        """Return a list of all recipients of the e-mail.

        Returns:
            list:
            A list of all recipients included on the To, CC, and BCC lists.
        """
        return self.to + self.bcc + self.cc



from django.db import models


class TestEvolveExtensionModel(models.Model):
    test_field = models.CharField(max_length=16)
    new_field = models.IntegerField(default=42)

    class Meta(object):
        app_label = 'evolve_tests'
        db_table = 'evolve_tests_testevolveextensionmodel'



"""A manager for tracking and working with integrations."""

import atexit
import logging
import threading
import weakref

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models.signals import post_delete, post_save

from djblets.cache.synchronizer import GenerationSynchronizer
from djblets.integrations.errors import (IntegrationAlreadyRegisteredError,
                                         IntegrationNotRegisteredError,
                                         IntegrationRegistrationError)


_integration_managers = weakref.WeakValueDictionary()


class IntegrationManager(object):
    """Manages integrations with third-party services.

    The manager keeps track of the integrations registered by extensions
    or other components of an application, providing the ability to
    register new ones, unregister existing ones, and list any that are
    currently enabled.

    It also manages the lookups of configurations for integrations, taking
    care to cache the lookups for any integrations and invalidate them when
    a configuration has been updated.

    Attributes:
        config_model (type):
            The model used to store configuration data. This is a subclass of
            :py:class:`djblets.integrations.models.BaseIntegrationConfig`.

        enabled (bool):
            The integration manager is enabled and can be used for registering,
            unregistering, and otherwise using integrations. If this is
            ``False``, then :py:meth:`shutdown` has been called, and it should
            be assumed that no integrations are registered or need to be
            unregistered.
    """

    def __init__(self, config_model):
        """Initialize the integration manager.

        Args:
            config_model (type):
                The model used to store configuration data. This must be a
                subclass of
                :py:class:`djblets.integrations.models.BaseIntegrationConfig`.
        """
        # Check that the Django environment is set up for integrations to
        # properly function.
        if 'djblets.integrations' not in settings.INSTALLED_APPS:
            raise ImproperlyConfigured(
                'IntegrationManager requires djblets.integrations to be '
                'listed in settings.INSTALLED_APPS.'
            )

        middleware = 'djblets.integrations.middleware.IntegrationsMiddleware'

        MIDDLEWARE = getattr(settings, 'MIDDLEWARE', None) or []

        if middleware not in MIDDLEWARE:
            raise ImproperlyConfigured(
                'IntegrationManager requires %s to be listed in '
                'settings.MIDDLEWARE'
                % middleware
            )

        self.config_model = config_model

        key = ('integrationmgr:%s.%s'
               % (self.__class__.__module__, self.__class__.__name__))

        self.enabled = True

        self._integration_classes = {}
        self._integration_configs = {}
        self._integration_instances = {}
        self._lock = threading.Lock()
        self._needs_recalc = False
        self._gen_sync = GenerationSynchronizer('%s:gen' % key)

        instance_id = id(self)
        _integration_managers[instance_id] = self

        # Listen for any config model saves/deletes, so we can mark whether
        # a reload of state is needed.
        dispatch_uid = '%s:%s' % (key, id(self))

        post_delete.connect(self._on_config_changes,
                            sender=config_model,
                            dispatch_uid=dispatch_uid)
        post_save.connect(self._on_config_changes,
                          sender=config_model,
                          dispatch_uid=dispatch_uid)

    def shutdown(self):
        """Shut down the integrations on this integration manager.

        This should be called when the integration manager and integrations
        will no longer be used. It will shut down every integration and
        unregister all integrations.
        """
        for integration in self.get_integrations():
            integration.disable_integration()

        self._integration_classes = {}
        self._integration_configs = {}
        self._integration_instances = {}
        self._needs_recalc = False

        try:
            del _integration_managers[id(self)]
        except KeyError:
            pass

        self.enabled = False

    def get_integration_classes(self):
        """Return all the integration classes that have been registered.

        This is not sorted in any particular order. It is up to the caller
        to determine the correct sorting order.

        Yields:
            type:
            The registered integration classes.
        """
        return self._integration_classes.values()

    def get_integration(self, integration_id):
        """Return an integration instance for a given ID.

        Args:
            integration_id (unicode):
                The integration ID that was registered.

        Returns:
            djblets.integrations.integration.Integration:
            The integration instance.

        Raises:
            djblets.integrations.errors.IntegrationNotRegisteredError:
                The integration class provided wasn't registered.
        """
        try:
            return self._integration_instances[integration_id]
        except KeyError:
            raise IntegrationNotRegisteredError(integration_id)

    def get_integrations(self):
        """Return all the integration instances.

        This is not sorted in any particular order. It is up to the caller
        to determine the correct sorting order.

        Yields:
            djblets.integrations.integration.Integration:
            The integration instances.
        """
        return self._integration_instances.values()

    def get_integration_configs(self, integration_cls=None, **filter_kwargs):
        """Return a list of saved integration configurations.

        By default, all configurations will be returned for all integrations,
        including configurations that are disabled. This can be filtered
        down by specifying an integration class and/or by filtering by fields
        in the model through keyword arguments.

        Each set of results for a unique combination of integration class and
        filter arguments will be cached locally, to speed up further lookups.
        This cache can be flushed using :py:meth:`clear_configs_cache` or
        :py:meth:`clear_all_configs_cache`, and will be automatically cleared
        when cnofigurations are added, updated, or removed.

        Args:
            integration_cls (type, optional):
                The integration class to filter by.

            **filter_kwargs (dict, optional):
                Keyword arguments to filter by. Each must match a field and
                value on the model.

        Returns:
            list of djblets.integrations.models.BaseIntegrationConfig:
            A list of saved integration configurations matching the query.
        """
        key = self._make_config_filter_cache_key(integration_cls,
                                                 **filter_kwargs)

        try:
            configs = self._integration_configs[key]
        except KeyError:
            queryset = self.config_model.objects.all()

            if integration_cls:
                queryset = queryset.filter(
                    integration_id=integration_cls.integration_id)

            if filter_kwargs:
                queryset = queryset.filter(**filter_kwargs)

            configs = list(queryset)
            self._integration_configs[key] = configs

        return configs

    def clear_configs_cache(self, integration_cls=None, **filter_kwargs):
        """Clear the configuration cache matching the given filters.

        This is used to clear a subset of the configs cache, matching the exact
        query arguments passed to a previous call to
        :py:meth:`get_integration_configs`.

        To clear the entire cache, use :py:meth:`clear_all_configs_cache`.

        Args:
            integration_cls (type, optional):
                The integration class for the filter.

            **filter_kwargs (dict, optional):
                Keyword arguments for the filter.
        """
        key = self._make_config_filter_cache_key(integration_cls,
                                                 **filter_kwargs)

        try:
            del self._integration_configs[key]
        except KeyError:
            pass

    def clear_all_configs_cache(self):
        """Clear the entire configuration cache.

        This will force all future lookups to re-query the database. To
        clear only a subset of the cache, use :py:meth:`clear_configs_cache`.
        """
        self._integration_configs = {}

    def is_expired(self):
        """Return whether the integration manager has expired state.

        Returns:
            bool:
            ``True`` if there's either expired configuration state or
            integrations that need their enabled state recalculated.
        """
        return self._needs_recalc or self._gen_sync.is_expired()

    def check_expired(self):
        """Check for and handle expired integration state.

        If the configurations of one or more integrations have been updated
        by another process, or there are new integrations registered that
        may need to be enabled, this method will reset the cache state and
        re-calculate the integrations to enable/disable.
        """
        if self.is_expired():
            # We're going to check the expiration, and then only lock if it's
            # expired. Following that, we'll check again.
            #
            # We do this in order to prevent locking unnecessarily, which could
            # impact performance or cause a problem if a thread is stuck.
            #
            # We're checking the expiration twice to prevent every blocked
            # thread from making its own attempt to reload the integrations
            # state the first thread holding the lock finishes.
            with self._lock:
                # Check again, since another thread may have already
                # reloaded.
                if self.is_expired():
                    self._gen_sync.refresh()
                    self.clear_all_configs_cache()
                    self._recalc_enabled_integrations()

    def register_integration_class(self, integration_cls):
        """Register a class for an integration.

        This will instantiate the integration and make it available for new
        configurations.

        Args:
            integration_cls (type):
                The integration class to register.

        Returns:
            djblets.integrations.integration.Integration:
            The new instance of the registered integration class.

        Raises:
            djblets.integrations.errors.IntegrationAlreadyRegisteredError:
                The integration class was already registered.

            djblets.integrations.errors.IntegrationConstructionError:
                Error initializing an instance of the integration. The
                integration will not be registered.
        """
        if not integration_cls.integration_id:
            # A pre-existing integration ID doesn't exist, so create one
            # based on the class path.
            integration_cls.integration_id = '%s.%s' % (
                integration_cls.__module__,
                integration_cls.__name__)

        integration_id = integration_cls.integration_id

        if integration_id in self._integration_classes:
            raise IntegrationAlreadyRegisteredError(integration_id)

        try:
            # We're going to instantiate the integration, but we won't
            # initialize it until later when we know that there are
            # configurations available.
            integration = integration_cls(self)
            self._integration_instances[integration_id] = integration
            self._integration_classes[integration_id] = integration_cls
        except Exception as e:
            # This should never happen, unless the subclass overrode
            # __init__.
            raise IntegrationRegistrationError(
                'Unexpected error when constructing integration %s: %s'
                % (integration_cls.__name__, e))

        # Flag that we need to recalculate the list of enabled integrations.
        # The next time a request is served, the middleware will perform the
        # recalculation. We do it this way instead of during registration in
        # order to cut back on the number of queries needed.
        self._needs_recalc = True

        return integration

    def unregister_integration_class(self, integration_cls):
        """Unregister a class for an integration.

        The integration instance will be shut down, and the integration will
        no longer be made available for any further configuration.

        If there is an error shutting down the integration, the output will
        be logged, but no error will be returned.

        Args:
            integration_cls (type):
                The integration class to unregister.

        Raises:
            djblets.integrations.errors.IntegrationNotRegisteredError:
                The integration class was not previously registered.
        """
        integration_id = integration_cls.integration_id

        if (not integration_id or
            integration_id not in self._integration_classes):
            raise IntegrationNotRegisteredError(integration_id)

        integration = self.get_integration(integration_id)

        if integration.enabled:
            try:
                integration.disable_integration()
            except Exception as e:
                logging.exception(
                    'Unexpected error when shutting down integration %r: %s',
                    integration_cls, e)

        del self._integration_classes[integration_id]
        del self._integration_instances[integration_id]

    def _recalc_enabled_integrations(self):
        """Recalculate the enabled states of all integrations.

        The list of enabled configurations for integrations will be queried
        from the database. Based on this, the desired enabled state of each
        integration will be calculated. Those that are disabled but have
        enabled configurations will be enabled, and those that are enabled
        but no longer have enabled configurations will be disabled.

        This allows us to keep memory requirements and event handling at a
        minimum for any integrations not currently in use.
        """
        enabled_integration_ids = set(
            self.config_model.objects
            .filter(enabled=True)
            .distinct()
            .values_list('integration_id', flat=True)
        )

        for integration in self.get_integrations():
            should_enable = (integration.integration_id in
                             enabled_integration_ids)

            if should_enable != integration.enabled:
                if should_enable:
                    integration.enable_integration()
                else:
                    integration.disable_integration()

        self._needs_recalc = False

    def _make_config_filter_cache_key(self, integration_cls, **filter_kwargs):
        """Return a cache key for a config query filter.

        Args:
            integration_cls (type):
                The integration class used for the query.

            **filter_kwargs (dict):
                The filter keyword arguments used for the query.

        Returns:
            unicode:
            The resulting cache key.
        """
        if integration_cls:
            return '%s:%s' % (integration_cls.integration_id, filter_kwargs)
        else:
            return '*:%s' % (filter_kwargs,)

    def _on_config_changes(self, **kwargs):
        """Handler for when configuration state changes.

        This will force the list of integrations to recalculate on this
        process and others when a configuration is created, saved, or deleted.

        Args:
            **kwargs (dict):
                Keyword arguments passed to the signal.
        """
        self._needs_recalc = True
        self._gen_sync.mark_updated()


def get_integration_managers():
    """Return all integration manager instances.


    This will return all the integration managers that have been constructed.
    The order is not guaranteed.

    Returns:
        list: The list of :py:class:`IntegrationManager` instances.
    """
    return list(_integration_managers.values())


def shutdown_integration_managers():
    """Shut down all integration managers.

    This is called automatically when the process exits, but can be run
    manually.
    """
    # Make a copy of the list as we iterate, since the integration managers
    # will be removing themselves from the list.
    for integration_manager in list(get_integration_managers()):
        integration_manager.shutdown()


# When the process ends, shut down the integrations on each manager. That will
# help work around bugs in Django where it attempts to work with garbage state
# being held onto by extensions.
atexit.register(shutdown_integration_managers)



"""Helpers for database evolutions."""

from django_evolution.mutations import BaseMutation


class FakeChangeFieldType(BaseMutation):
    """
    Changes the type of the field to a similar type.
    This is intended only when the new type is really a version of the
    old type, such as a subclass of that Field object. The two fields
    should be compatible or there could be migration issues.
    """
    def __init__(self, model_name, field_name, new_type):
        self.model_name = model_name
        self.field_name = field_name
        self.new_type = new_type

    def __repr__(self):
        return "FakeChangeFieldType('%s', '%s', '%s')" % \
            (self.model_name, self.field_name, self.new_type)

    def simulate(self, app_label, proj_sig):
        app_sig = proj_sig[app_label]
        model_sig = app_sig[self.model_name]
        field_dict = model_sig['fields']
        field_sig = field_dict[self.field_name]

        field_sig['field_type'] = self.new_type

    def mutate(self, app_label, proj_sig):
        # We can just call simulate, since it does the same thing.
        # We're not actually generating SQL, but rather tricking
        # Django Evolution.
        self.simulate(app_label, proj_sig)
        return ""



"""Unit tests for the Root Resource."""

from django.core.exceptions import ImproperlyConfigured
from django.test.client import RequestFactory
from django.test.utils import override_settings

from djblets.extensions.manager import ExtensionManager
from djblets.extensions.resources import ExtensionResource
from djblets.testing.testcases import TestCase
from djblets.webapi.resources import RootResource, WebAPIResource


# Will appear in the URI templates list under the names 'mock' and 'mocks'.
class MockResource(WebAPIResource):
    name = 'mock'
    uri_object_key = 'mock_id'


# Will cause a conflict with MockResource when added to the URI templates list.
class DuplicateMockResource(MockResource):
    uri_name = 'duplicatemocks'


# Will appear in the URI templates list under the names 'test_mock' and
# 'test_mocks'.
class OtherMockResource(MockResource):
    uri_template_name = 'test_mock'


# Will be excluded from the URI templates list.
class ExcludedResource(WebAPIResource):
    name = 'exclude'
    uri_template_name = None
    uri_object_key = 'exclude_id'


# Will have its item resource excluded but list resource included in the
# URI templates list.
class IncludedListResource(WebAPIResource):
    name = 'test_list'
    uri_template_name = None
    uri_template_name_plural = 'lists'


class RootResourceTests(TestCase):
    """Unit tests for RootResource.

    Version Added:
        3.1.0
    """

    def setUp(self):
        """Setup for the RootResource unit tests."""
        super().setUp()

        self.root_res = RootResource(
            [
                MockResource(),
                OtherMockResource(),
                ExcludedResource(),
                IncludedListResource(),
            ]
        )
        self.request = RequestFactory().get('/')

    def test_get_uri_templates_uses_uri_template_names(self):
        """Testing RootResource.get_uri_templates uses the uri_template_name
        and uri_template_name_plural for resources instead of name and
        plural_name
        """
        uri_templates = self.root_res.get_uri_templates(self.request)

        self.assertEqual(uri_templates['mock'],
                         'http://testserver/mocks/{mock_id}/')
        self.assertEqual(uri_templates['mocks'], 'http://testserver/mocks/')

        self.assertEqual(uri_templates['test_mock'],
                         'http://testserver/mocks/{mock_id}/')
        self.assertEqual(uri_templates['test_mocks'],
                         'http://testserver/mocks/')

        self.assertEqual(uri_templates['lists'],
                         'http://testserver/test-lists/')

    def test_get_uri_templates_skips_with_none(self):
        """Testing RootResource.get_uri_templates skips item resources that
        have their uri_template_name and list resources that have their
        uri_template_plural_name set to None
        """
        uri_templates = self.root_res.get_uri_templates(self.request)

        self.assertEqual(uri_templates['mock'],
                         'http://testserver/mocks/{mock_id}/')
        self.assertEqual(uri_templates['mocks'], 'http://testserver/mocks/')

        self.assertEqual(uri_templates['test_mocks'],
                         'http://testserver/mocks/')
        self.assertEqual(uri_templates['test_mock'],
                         'http://testserver/mocks/{mock_id}/')

        self.assertEqual(uri_templates['lists'],
                         'http://testserver/test-lists/')
        self.assertNotIn('test_list', uri_templates)

        self.assertNotIn('exclude', uri_templates)
        self.assertNotIn('excludes', uri_templates)

    @override_settings(DEBUG=False)
    def test_get_uri_templates_must_be_unique(self):
        """Testing RootResource.get_uri_templates logs an error when multiple
        URI templates are mapped to the same name in production mode
        """
        self.root_res = RootResource([MockResource(), DuplicateMockResource()])
        expected_message = (
            'More than one URI template was mapped to the "mocks" name: '
            'http://testserver/mocks/, http://testserver/duplicatemocks/. '
            'Only the first one will be included in the URI templates list. '
            'To include the other URI templates, they must be mapped to a '
            'unique name by setting each resource\'s uri_template_name '
            'property.'
        )

        with self.assertLogs() as logs:
            uri_templates = self.root_res.get_uri_templates(self.request)

            self.assertEqual(logs.records[0].getMessage(), expected_message)
            self.assertIn('mocks', uri_templates)
            self.assertEqual(uri_templates['mocks'],
                             'http://testserver/mocks/')

    @override_settings(DEBUG=True)
    def test_get_uri_templates_must_be_unique_debug(self):
        """Testing RootResource.get_uri_templates raises an error when multiple
        URI templates are mapped to the same name in debug mode
        """
        self.root_res = RootResource([MockResource(), DuplicateMockResource()])
        expected_message = (
            'More than one URI template was mapped to the "mocks" name: '
            'http://testserver/mocks/, http://testserver/duplicatemocks/. '
            'Each URI template must be mapped to a unique URI template '
            'name in order to be included in the URI templates list. This can '
            'be set through the uri_template_name property.'
        )

        with self.assertRaisesMessage(ImproperlyConfigured, expected_message):
            self.root_res.get_uri_templates(self.request)


class RootResourceTemplateRegistrationTests(TestCase):
    """Unit tests for the (un)registration of templates in RootResource."""

    def setUp(self):
        super(RootResourceTemplateRegistrationTests, self).setUp()

        self.ext_mgr = ExtensionManager('')
        self.ext_res = ExtensionResource(self.ext_mgr)
        self.root_res = RootResource([self.ext_res])
        self.root_res._registered_uri_templates = {
            self.ext_res: {
                'extensions': 'http://localhost:8080/api/extensions/'
            },
            None: {
                'extensions': 'http://localhost:8080/api/extensions/none/'
            },
        }

    def test_register_uri_template_without_relative_resource(self):
        """Testing register_uri_templates without a relative resource"""
        self.root_res.register_uri_template(name='key', relative_path='value')
        actual_result = self.root_res._registered_uri_templates[None]
        self.assertEqual(actual_result, {
            'extensions': 'http://localhost:8080/api/extensions/none/',
            'key': 'value',
        })

    def test_register_uri_template_with_relative_resource(self):
        """Testing register_uri_templates with a relative resource"""
        mock_extension_resource = ExtensionResource(self.ext_mgr)
        self.root_res.register_uri_template(
            name='key',
            relative_path='value',
            relative_resource=mock_extension_resource)
        actual_result = self.root_res._registered_uri_templates[
            mock_extension_resource]
        self.assertEqual(actual_result, {'key': 'value'})

    def test_register_uri_template_clears_uri_template_cache(self):
        """Testing register_uri_templates clears the URI template cache"""
        self.root_res._uri_templates = {
            'key1': 'value1',
            'key2': 'value2',
        }
        self.root_res.register_uri_template(
            'extension_name', 'some/relative/path/')
        self.assertEqual(self.root_res._uri_templates, {})

    def test_register_uri_template_overwrites_existing_uri_template(self):
        """Testing register_uri_templates overwrites existing uri templates and
        logs a message saying so
        """
        expected_message = ('The extensions resource is already mapped to the '
                            'following URI template: '
                            'http://localhost:8080/api/extensions/none/. This '
                            'will be overwritten by the new URI template: '
                            'http://localhost:8080/api/different/.')

        with self.assertLogs(level='DEBUG') as logs:
            self.root_res.register_uri_template(
                name='extensions',
                relative_path='http://localhost:8080/api/different/')
            actual_result = self.root_res._registered_uri_templates[None]

            self.assertEqual(logs.records[0].getMessage(), expected_message)
            self.assertEqual(actual_result, {
                'extensions': 'http://localhost:8080/api/different/',
            })

    def test_unregister_uri_template_without_relative_resource(self):
        """Testing unregister_uri_template without a relative resource"""
        self.root_res.unregister_uri_template('extensions')
        self.assertFalse(self.root_res._registered_uri_templates[None])

    def test_unregister_uri_template_with_relative_resource(self):
        """Testing unregister_uri_template with a relative resource"""
        self.root_res.unregister_uri_template('extensions', self.ext_res)
        self.assertFalse(self.root_res._registered_uri_templates[
            self.ext_res])

    def test_unregister_uri_template_clears_uri_template_cache(self):
        """Testing unregister_uri_templates clears the URI template cache"""
        self.root_res._uri_templates = {
            'key1': 'value1',
            'key2': 'value2',
        }
        self.root_res.unregister_uri_template('extension-1')
        self.assertEqual(self.root_res._uri_templates, {})


"""A form for working with key/value stores."""

from django import forms


class KeyValueForm(forms.Form):
    """A form for working with key/value stores.

    Typical forms are built to work either with database models or with
    entirely custom objects, but are less useful when working with
    dictionaries, caches, generic object attributes, or other objects that work
    as key/value stores.

    This form provides a standard way of loading fields from a key/value
    store and saving back out to one.

    By default, it assumes it's working with something that behaves like a
    dictionary (containing a ``get()`` method and a ``__getitem__`` operator).
    This can be overridden by providing implementations of
    :py:meth:`get_key_value` and :py:meth:`set_key_value`.

    Values for specific keys can be specially serialized/deserialized by
    providing :samp:`serialize_{keyname}_field()` and
    :samp:`deserialize_{keyname}_field()` functions. These take a value and are
    expected to return a serializable JSON value or a deserialized value,
    respectively.

    It's also makes it easy to implement saving behavior for the object
    by overriding :py:meth:`save_instance`.

    There's support for dynamically marking fields as disabled and specifying
    the reason it's disabled, for display in the form. This can be done by
    setting :py:attr:`disabled_fields` and :py:attr:`disabled_reasons`, and
    using the latter in a template.

    Fields can be blacklisted from loading or saving by setting
    :py:attr:`Meta.load_blacklist` and :py:attr:`Meta.save_blacklist`,
    respectively to a list or tuple of field names. Unless specified otherwise,
    the loading blacklist will default to the same fields in the save
    blacklist.

    Args:
        disabled_fields (dict):
            A dictionary of field names to booleans indicating fields that
            should be disabled (if their values are set to ``True``). Those
            fields will then be disabled in the HTML (by setting the
            ``disabled`` attribute on the field).

        disabled_reasons (dict):
            A dictionary of field names to strings describing the reason a
            particular field is disabled. These are not required, and are
            not automatically outputted in any form, but may be useful to
            templates.

        instance (object):
            The instance being loaded and saved. This can be ``None`` (the
            default) when not yet working on an instance (but in that case,
            :py:meth:`create_instance` must be defined).

    Example:
        With custom field serialization:

        .. code-block:: python

            class MyForm(KeyValueForm):
                book = forms.ModelChoiceField(queryset=Book.objects.all())

                def serialize_book_field(self, value):
                    return {
                        'id': book.pk,
                        'title': book.title,
                    }

                def deserialize_book_field(self, value):
                    return Book.objects.get(pk=value['id'])
    """

    #: The list of CSS bundle names to include on the page.
    css_bundle_names = []

    #: The list of JavaScript bundle names to include on the page.
    js_bundle_names = []

    def __init__(self, data=None, files=None, instance=None, *args, **kwargs):
        """Initialize the form.

        Args:
            data (dict, optional):
                Data for the form.

            files (dict, optional):
                File uploads for the form.

            instance (object, optional):
                The existing instance being loaded from, if any.

            *args (tuple):
                Positional arguments for the form.

            **kwargs (dict):
                Keyword arguments for the form.
        """
        super(KeyValueForm, self).__init__(data=data, files=files,
                                           *args, **kwargs)

        self.instance = instance
        self.disabled_fields = {}
        self.disabled_reasons = {}

        self.load()

    def load(self):
        """Load form fields from the instance.

        If an instance was passed to the form, any values found in that
        instance will be set as the initial data for the form. If an instance
        was not passed, then the fields will be left as their default values.

        This also updates the disabled status of any fields marked as
        disabled in the :py:attr:`disabled_fields` attribute.
        """
        load_blacklist = self.get_load_blacklist()
        disabled_fields = set(self.disabled_fields)

        for field_name, field in self.fields.items():
            if self.instance is not None and field_name not in load_blacklist:
                value = self.get_key_value(field_name, default=field.initial)
                deserialize_func = getattr(self,
                                           'deserialize_%s_field' % field_name,
                                           None)

                if deserialize_func is not None:
                    value = deserialize_func(value)

                field.initial = value

            if field_name in disabled_fields:
                field.widget.attrs['disabled'] = 'disabled'

    def save(self, commit=True, extra_save_blacklist=[]):
        """Save form fields back to the instance.

        This will save the values of any fields not in the blacklist out to
        the instance.

        If the instance doesn't yet exist, it will be created first through a
        call to :py:meth:`create_instance`.

        Args:
            commit (boolean, optional):
                Whether to save the instance after setting all the fields.
                Defaults to ``True`` (though this will do nothing if
                :py:meth:`save_instance` is not overridden).

            extra_save_blacklist (list, optional):
                Additional fields that should not be saved from the form.

        Raises:
            ValueError:
                The form couldn't be saved due to errors.
        """
        if self.errors:
            raise ValueError('The form could not be saved due to one or more '
                             'errors.')

        if self.instance is None:
            self.instance = self.create_instance()

        blacklist = set(extra_save_blacklist) | set(self.get_save_blacklist())

        for key, value in self.cleaned_data.items():
            if key not in blacklist:
                serialize_func = getattr(self, 'serialize_%s_field' % key,
                                         None)

                if serialize_func is not None:
                    value = serialize_func(value)

                self.set_key_value(key, value)

        if commit:
            self.save_instance()

        return self.instance

    def get_key_value(self, key, default=None):
        """Return the value for a key in the instance.

        This defaults to calling a ``get()`` method on the instance,
        passing in the values for ``key`` and ``default`` as positional
        arguments.

        This can be overridden to change how values are loaded from the
        instance.

        Args:
            key (unicode):
                The key to fetch from the instance. This will be a field
                name in the form.

            default (object, optional):
                The default value, from the field's initial value.

        Returns:
            object:
            The value from the instance.
        """
        return self.instance.get(key, default)

    def set_key_value(self, key, value):
        """Set a value in the instance.

        This defaults to calling the ``[]`` operator (as in ``instance[key]``)
        to set the value.

        This can be overridden to change how values are set in the instance.

        Args:
            key (unicode):
                The key in the instance where the value should be stored.

            value (object):
                The value to store in the instance.
        """
        self.instance[key] = value

    def create_instance(self):
        """Create a new instance.

        If an instance was not provided to the form, this will be called
        prior to saving. It must return an instance that can have new values
        set.

        If not implemented, and no instance is passed to the form, this
        will raise a :py:exc:`NotImplementedError` when called.

        Returns:
            object:
            The resulting instance to populate.

        Raises:
            NotImplementedError:
                The method was not overridden by a subclass, but was called
                due to an instance not being passed to the form.
        """
        raise NotImplementedError('%r must implement create_instance()'
                                  % self.__class__)

    def save_instance(self):
        """Save the instance.

        By default, this doesn't do anything. Subclasses can override it to
        provide saving functionality for the instance.
        """
        pass

    def get_load_blacklist(self):
        """Return the field blacklist for loading.

        Any field names returned from here will not be loaded from the
        instance.

        If the subclass has a :py:attr:`Meta.load_blacklist` attribute, it will
        be returned. If not, but it has a :py:attr:`Meta.save_blacklist`
        attribute, that will be returned. Otherwise, an empty list will be
        returned.

        This can be overridden to provide more specialized behavior.

        Returns:
            list:
            The field names to blacklist from loading from the instance.
        """
        if hasattr(self, 'Meta'):
            return getattr(self.Meta, 'load_blacklist',
                           self.get_save_blacklist())

        return []

    def get_save_blacklist(self):
        """Return the field blacklist for saving.

        Any field names returned from here will not be saved to the
        instance.

        If the subclass has a :py:attr:`Meta.save_blacklist` attribute, it will
        be returned. Otherwise, an empty list will be returned.

        This can be overridden to provide more specialized behavior.

        Returns:
            list:
            The field names to blacklist from saving to the instance.
        """
        if hasattr(self, 'Meta'):
            return getattr(self.Meta, 'save_blacklist', [])

        return []



from django.http import HttpResponse
from django.urls import include, path


def dummy_view(request):
    return HttpResponse('')


urlpatterns = [
    path('', dummy_view, name='test-url-name'),
    path('admin/extensions/', include('djblets.extensions.urls')),
    path('auth/', include('djblets.auth.urls')),
]



"""A specialization of pipeline's LessCompiler."""

import logging
import os
import subprocess
from datetime import datetime

from django.utils.functional import cached_property
from pipeline.compilers.less import LessCompiler as PipelineLessCompiler
from pipeline.conf import settings as pipeline_settings


class LessCompiler(PipelineLessCompiler):
    """A LESS compiler which doesn't redirect the output.

    Pipeline's LESS compiler outputs to stdout and redirects that output to the
    target file, because of some bug in old versions of lessc. We don't want to
    do that, because it interferes with the generation of sourcemaps.

    To use this, add the path to this class to
    ``settings.PIPELINE['COMPILERS']``.

    Version Changed:
        2.2:
        Added in-memory caching of import lists and file information to
        speed up performance of compilations.
    """

    #: Cache expiration time for the list of outdated timestamp states.
    #:
    #: Version Added:
    #:     2.2
    _MTIME_CACHE_STALE_SECS = 10

    #: Cache mapping files to lists of imported files.
    #:
    #: Version Added:
    #:     2.2
    #:
    #: Type:
    #:     dict
    _import_files_cache = {}

    #: Cache of filenames to mtimes.
    #:
    #: Version Added:
    #:     2.2
    #:
    #: Type:
    #:     dict
    _mtime_cache = {}

    #: The latest updated timestamp for the mtime cache.
    #:
    #: Version Added:
    #:     2.2
    #:
    #: Type:
    #:     datetime.datetime
    _mtime_cache_stamp = None

    @classmethod
    def clear_caches(cls):
        """Clear the caches used for faster compilation."""
        cls._import_files_cache.clear()
        cls._mtime_cache.clear()
        cls._mtime_cache_stamp = None

    @cached_property
    def less_imports_command(self):
        """Command line and arguments for less-imports.js.

        Type:
            unicode
        """
        return [
            os.path.join(os.path.dirname(__file__), 'less-imports.js'),
        ] + list(pipeline_settings.LESS_ARGUMENTS)

    def compile_file(self, infile, outfile, outdated=False, force=False):
        """Compile a given file.

        Args:
            infile (unicode):
                The input filename.

            outfile (unicode):
                The output filename.

            outdated (bool):
                Whether the file is outdated.

            force (bool):
                Whether the compilation should proceed, regardless of whether
                the file is out of date.
        """
        if outdated or force:
            self.execute_command((
                pipeline_settings.LESS_BINARY,
                pipeline_settings.LESS_ARGUMENTS,
                infile,
                outfile
            ))

    def is_outdated(self, infile, outfile):
        """Return whether a given file is outdated.

        This will compare the timestamps of the ``infile`` and ``outfile``, but
        also compare the timestamp of ``outfile`` against any import
        depedencies that exist.

        Args:
            infile (unicode):
                The path to the input file.

            outfile (unicode):
                The path to the output file.

        Returns:
            bool:
            True if the file is out of date and needs to be recompiled.
        """
        # If the file itself is outdated, or we want to always treat all files
        # as outdated, don't bother to dive into dependencies.
        if pipeline_settings.get('_DJBLETS_LESS_ALWAYS_REBUILD', False):
            return True

        outfile_mtime = self._get_mtime(outfile)

        if self._is_file_timestamp_outdated(infile, outfile_mtime):
            return True

        try:
            # If the file is in the cache, then start by trusting that the
            # imports haven't changed. We know this file is up-to-date (since
            # we checked above). We'll next be checking each dependency
            # (potentially invalidating all or part of the file cache, if we
            # find one that is out-of-date).
            files = self._import_files_cache[infile]
        except KeyError:
            # Parse out a deep list of the imported files and check those.
            files = self._get_import_files(infile)

            if files is None:
                # We've already logged. Consider this file outdated.
                return True

            self._import_files_cache[infile] = files

        outdated_deps = [
            dep
            for dep in files
            if self._is_file_timestamp_outdated(dep, outfile_mtime)
        ]

        if outdated_deps:
            # Any of the dependencies we found will have already invalidated
            # any file import caches referencing them. Just log and return.
            logging.info('Recompiling %s because one or more dependencies '
                         '(%s) are out of date.',
                         infile, ', '.join(outdated_deps))

            return True

        return False

    def _is_file_timestamp_outdated(self, filename, compare_mtime):
        """Return whether a file's timestamp is outdated.

        If the file is outdated, any cached entries referencing the file will
        be cleared out so that they can be refetched on the next compilation
        check.

        Args:
            filename (unicode):
                The filename to look up.

            compare_mtime (int):
                A pre-fetched mtime to compare the file's mtime to.

        Returns:
            bool:
            ``True`` if the file is outdated. ``False`` if it is up-to-date.
        """
        if compare_mtime is None:
            outdated = True
        else:
            mtime = self._get_mtime(filename)
            outdated = (mtime is None or mtime > compare_mtime)

        if outdated:
            # Since this can run concurrently with other operations, we need
            # to take care not to assume the presence of any keys.
            import_files_cache = self._import_files_cache
            import_files_cache.pop(filename, None)

            for key, files in list(import_files_cache.items()):
                if filename in files:
                    import_files_cache.pop(key, None)

        return outdated

    def _get_import_files(self, filename):
        """Return the files imported when compiling a LessCSS file.

        This will return all files that are ultimately imported by a file,
        recursively. That means it will include files imported by files
        imported by ``filename``.

        Args:
            filename (unicode):
                The filename to return imports for.

        Returns:
            set of unicode:
            The filenames imported, or ``None`` if there was an error
            parsing any of the files.
        """
        p = subprocess.Popen(self.less_imports_command + [filename],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        output, errors = p.communicate()

        rc = p.wait()

        if rc == 0:
            files = set(output.decode('utf-8').splitlines())
        else:
            logging.error('Error reading LessCSS imports from %s: %s',
                          filename, errors.decode('utf-8'))
            files = None

        return files

    def _get_mtime(self, filename):
        """Return the mtime for a file.

        This will utilize a cache that remains valid for only a short period
        of time, cutting back on redundant stat operations on the filesystem
        during the compilation phases.

        Args:
            filename (unicode):
                The filename to look up.

        Returns:
            int:
            The file's mtime, or ``None`` if the file was not found.
        """
        cls = type(self)
        cache = cls._mtime_cache
        stamp = cls._mtime_cache_stamp
        now = datetime.now()

        if not stamp or (now - stamp).seconds > self._MTIME_CACHE_STALE_SECS:
            cache.clear()
            cls._mtime_cache_stamp = now

        try:
            mtime = cache[filename]
        except KeyError:
            try:
                mtime = os.path.getmtime(filename)
            except OSError:
                mtime = None

            cache[filename] = mtime

        return mtime



"""Unit tests for djblets.privacy.consent.forms.ConsentConfigPageFormMixin."""

from django.contrib.auth.models import User
from django.contrib.messages import get_messages
from django.contrib.messages.middleware import MessageMiddleware
from django.contrib.sessions.middleware import SessionMiddleware
from django.http import HttpResponse
from django.test.client import RequestFactory

from djblets.configforms.forms import ConfigPageForm
from djblets.configforms.pages import ConfigPage
from djblets.configforms.views import ConfigPagesView
from djblets.privacy.consent import (BaseConsentRequirement, Consent,
                                     get_consent_tracker,
                                     get_consent_requirements_registry)
from djblets.privacy.consent.forms import ConsentConfigPageFormMixin
from djblets.privacy.tests.testcases import ConsentTestCase


class MyForm(ConsentConfigPageFormMixin, ConfigPageForm):
    def get_extra_consent_data(self):
        return {
            'test': True,
        }


class MyPage(ConfigPage):
    form_classes = [MyForm]


class MyConsentRequirement1(BaseConsentRequirement):
    requirement_id = 'my-requirement-1'
    name = 'My Requirement 1'
    summary = 'We would like to use this thing'
    intent_description = 'We need this for testing.'
    data_use_description = 'Sending all the things.'
    icons = {
        '1x': '/static/consent.png',
        '2x': '/static/consent@2x.png',
    }


class MyConsentRequirement2(BaseConsentRequirement):
    requirement_id = 'my-requirement-2'
    name = 'My Requirement 2'
    summary = 'We would also like this'
    intent_description = 'We need this for dancing.'
    data_use_description = 'Dancing all the things.'


class ConsentConfigPageFormMixinTests(ConsentTestCase):
    """Unit tests for ConsentConfigPageFormMixinTests."""

    def setUp(self):
        super(ConsentConfigPageFormMixinTests, self).setUp()

        self.registry = get_consent_requirements_registry()

        self.consent_requirement_1 = MyConsentRequirement1()
        self.registry.register(self.consent_requirement_1)

        self.consent_requirement_2 = MyConsentRequirement2()
        self.registry.register(self.consent_requirement_2)

        self.user = User.objects.create(username='test-user')

        self.request = RequestFactory().get('/consent/')
        self.request.user = self.user

        # Enable support for messages.
        middleware = SessionMiddleware(
            MessageMiddleware(
                lambda request: HttpResponse('')))
        middleware(self.request)

        self.page = MyPage(config_view=ConfigPagesView(),
                           request=self.request,
                           user=self.user)

    def test_init(self):
        """Testing ConsentConfigPageFormMixin.__init__ defines field"""
        get_consent_tracker().record_consent_data(
            self.user,
            self.consent_requirement_2.build_consent_data(granted=False))

        form = MyForm(page=self.page,
                      request=self.request,
                      user=self.user)
        self.assertIn('consent', form.fields)

        field = form.fields['consent']
        self.assertEqual(field.initial, [Consent.UNSET, Consent.DENIED])
        self.assertEqual(
            field.consent_requirements,
            [self.consent_requirement_1, self.consent_requirement_2])

        for subfield in field.fields:
            self.assertEqual(subfield.consent_source,
                             'http://testserver/consent/')
            self.assertEqual(
                subfield.extra_consent_data,
                {
                    'test': True,
                })

    def test_save(self):
        """Testing ConsentConfigPageFormMixin.save"""
        form = MyForm(
            page=self.page,
            request=self.request,
            user=self.user,
            data={
                'consent_my-requirement-1_choice': 'allow',
                'consent_my-requirement-2_choice': 'block',
            })
        self.assertTrue(form.is_valid())

        self.assertEqual(self.consent_requirement_1.get_consent(self.user),
                         Consent.UNSET)
        self.assertEqual(self.consent_requirement_2.get_consent(self.user),
                         Consent.UNSET)

        form.save()

        self.assertEqual(self.consent_requirement_1.get_consent(self.user),
                         Consent.GRANTED)
        self.assertEqual(self.consent_requirement_2.get_consent(self.user),
                         Consent.DENIED)

        # Check that the success message appeared.
        messages = list(get_messages(self.request))
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0].message,
                         'Your choices have been saved. You can make '
                         'changes to these at any time.')



"""Unit tests for djblets.extensions.manager.ExtensionManager."""

import errno
import logging
import os
import shutil
import tempfile
import threading
import time

import kgb
from django.conf import settings
from django.db import connection
from django.template import Context, Template, TemplateSyntaxError
from django.test.utils import override_settings
from django_evolution.evolve import Evolver
from pipeline.conf import settings as pipeline_settings

from djblets.extensions.errors import EnablingExtensionError
from djblets.extensions.extension import Extension
from djblets.extensions.hooks import URLHook
from djblets.extensions.manager import (ExtensionManager,
                                        get_extension_managers,
                                        logger as manager_logger)
from djblets.extensions.testing import ExtensionTestCaseMixin
from djblets.testing.testcases import TestCase


class MyTestExtension(Extension):
    # We set this to djblets.extensions to avoid an issue unique to tests
    # involving multiple ExtensionManagers. They'd both try wrapping
    # INSTALLED_APPS, and would try adding the same default app for the
    # extension upon load. While doing that, they'd assert that the extension
    # wasn't in INSTALLED_APPS, and that'd fail. The issue should only occur in
    # this test, or in a case of an app with multiple extension managers
    # loading the same list of apps.
    apps = [
        'djblets.extensions',
    ]


class HooksTestExtension(MyTestExtension):
    def initialize(self):
        self.url_hook = URLHook(self, ())


class ExtensionManagerTests(kgb.SpyAgency, ExtensionTestCaseMixin, TestCase):
    """Unit tests for djblets.extensions.manager.ExtensionManager."""

    @classmethod
    def setUpClass(cls):
        connection.disable_constraint_checking()
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        connection.enable_constraint_checking()

    def tearDown(self):
        super(ExtensionManagerTests, self).tearDown()

        self.assertEqual(len(URLHook.hooks), 0)

    def test_added_to_extension_managers(self):
        """Testing ExtensionManager registration"""
        manager = ExtensionManager('unit-tests')
        self.assertIn(manager, get_extension_managers())

    def test_get_enabled_extensions_returns_empty(self):
        """Testing ExtensionManager.get_enabled_extensions with no
        extensions
        """
        manager = ExtensionManager('unit-tests')
        manager.load()

        self.assertEqual(len(manager.get_enabled_extensions()), 0)

    def test_load(self):
        """Testing ExtensionManager.get_installed_extensions with loaded
        extensions
        """
        self.setup_extension(MyTestExtension, enable=False)

        self.assertEqual(self.extension_mgr.get_installed_extensions(),
                         [MyTestExtension])
        self.assertEqual(len(self.extension_mgr.get_enabled_extensions()), 0)
        self.assertTrue(hasattr(MyTestExtension, 'info'))
        self.assertEqual(MyTestExtension.info.name,
                         self.extension_package_name)
        self.assertTrue(hasattr(MyTestExtension, 'registration'))
        self.assertEqual(MyTestExtension.registration.name,
                         self.extension_package_name)

    def test_load_full_reload_hooks(self):
        """Testing ExtensionManager.load with full_reload=True"""
        # NOTE: We're not using HooksTestExtension here, because we want to
        #       manually set up a hook and not have to return once we
        #       perform a reload.
        extension = self.setup_extension(MyTestExtension)

        self.assertEqual(len(self.extension_mgr.get_installed_extensions()), 1)
        self.assertEqual(len(self.extension_mgr.get_enabled_extensions()), 1)

        URLHook(extension, ())
        self.assertEqual(len(URLHook.hooks), 1)
        self.assertEqual(URLHook.hooks[0].extension, extension)

        with self.scanned_extensions([MyTestExtension]):
            self.extension_mgr.load(full_reload=True)

        self.assertEqual(len(URLHook.hooks), 0)

    def test_load_concurrent_threads(self):
        """Testing ExtensionManager.load with concurrent threads"""
        # There are a number of things that could go wrong both during
        # uninitialization and during initialization of extensions, if
        # two threads attempt to reload at the same time and locking isn't
        # properly implemented.
        #
        # Extension uninit could be called twice, resulting in one thread
        # attempting to access state that's already been destroyed. We
        # could end up hitting:
        #
        #     "Extension's installed app <app> is missing a ref count."
        #     "'<Extension>' object has no attribute 'info'."
        #
        # (Without locking, we end up hitting the latter in this test.)
        #
        # If an extension is being initialized twice simultaneously, then
        # it can hit other errors. An easy one to hit is this assertion:
        #
        #     assert extension_id not in self._extension_instances
        #
        # With proper locking, these issues don't come up. That's what
        # this test case is attempting to check for.
        extension_mgr = self.extension_mgr

        # Enable one extension. This extension's state will get a bit messed
        # up if the thread locking fails. We only need one to trigger this.
        self.setup_extension(MyTestExtension)

        self.spy_on(extension_mgr._load_extensions)
        self._spy_sleep_and_call(extension_mgr._init_extension)
        self._spy_sleep_and_call(extension_mgr._uninit_extension)

        with self.scanned_extensions([MyTestExtension]):
            self._run_thread_test(lambda: extension_mgr.load(full_reload=True))

        self.assertSpyCallCount(extension_mgr._load_extensions, 2)
        self.assertSpyCallCount(extension_mgr._uninit_extension, 2)
        self.assertSpyCallCount(extension_mgr._init_extension, 2)
        self.assertEqual(self.exceptions, [])

    @override_settings(EXTENSIONS_ENABLED_BY_DEFAULT=[
        'djblets.extensions.tests.test_extension_manager.MyTestExtension',
    ])
    def test_load_with_enabled_by_default(self):
        """Testing ExtensionManager.load with
        settings.EXTENSIONS_ENABLED_BY_DEFAULT
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(MyTestExtension, enable=False)

        with self.scanned_extensions([MyTestExtension]):
            extension_mgr.load()

        self.assertEqual(extension_mgr.get_installed_extensions(),
                         [MyTestExtension])

        enabled_extensions = extension_mgr.get_enabled_extensions()
        self.assertEqual(len(enabled_extensions), 1)
        self.assertIsInstance(enabled_extensions[0], MyTestExtension)

        self.assertTrue(hasattr(MyTestExtension, 'info'))
        self.assertEqual(MyTestExtension.info.name,
                         self.extension_package_name)
        self.assertIsNotNone(MyTestExtension.instance)
        self.assertTrue(hasattr(MyTestExtension, 'registration'))
        self.assertEqual(MyTestExtension.registration.name,
                         self.extension_package_name)
        self.assertTrue(MyTestExtension.registration.enabled)

    def test_load_with_admin_site(self):
        """Testing ExtensionManager.load with has_admin_site=True"""
        class AdminSiteMyTestExtension(HooksTestExtension):
            has_admin_site = True

        manager = self.extension_mgr

        @self.spy_for(manager._init_admin_site)
        def _init_admin_site(_self, *args, **kwargs):
            self.assertIsNotNone(AdminSiteMyTestExtension.instance)
            self.assertIsInstance(AdminSiteMyTestExtension.instance,
                                  AdminSiteMyTestExtension)

            return manager._init_admin_site.call_original(*args, **kwargs)

        extension = self.setup_extension(AdminSiteMyTestExtension)

        self.assertEqual(manager.get_installed_extensions(),
                         [AdminSiteMyTestExtension])
        self.assertEqual(manager.get_enabled_extensions(), [extension])
        self.assertIs(AdminSiteMyTestExtension.instance, extension)

        self.assertIsNotNone(extension.admin_site)
        self.assertEqual(extension.admin_site.name, 'djblets.extensions.tests')

    def test_load_with_admin_site_failure(self):
        """Testing ExtensionManager.load with extension error setting up
        admin site
        """
        class AdminSiteMyTestExtension(HooksTestExtension):
            has_admin_site = True

        extension = self.setup_extension(AdminSiteMyTestExtension)
        manager = self.extension_mgr

        self.assertEqual(manager.get_installed_extensions(),
                         [AdminSiteMyTestExtension])
        self.assertEqual(manager.get_enabled_extensions(), [extension])
        self.assertIs(AdminSiteMyTestExtension.instance, extension)
        self.assertEqual(len(URLHook.hooks), 1)

        @self.spy_for(manager._init_admin_site)
        def _init_admin_site(*args, **kwargs):
            raise Exception('admin site failed!')

        with self.scanned_extensions([AdminSiteMyTestExtension]):
            self.extension_mgr.load(full_reload=True)

        self.assertEqual(manager.get_installed_extensions(),
                         [AdminSiteMyTestExtension])
        self.assertEqual(manager.get_enabled_extensions(), [])
        self.assertIsNone(AdminSiteMyTestExtension.instance)
        self.assertEqual(len(URLHook.hooks), 0)
        self.assertTrue(manager._load_errors[extension.id].startswith(
            'admin site failed!\n\nTraceback'))

    def test_load_with_admin_urls_failure(self):
        """Testing ExtensionManager.load with extension error setting up
        admin URLs
        """
        extension = self.setup_extension(HooksTestExtension)
        manager = self.extension_mgr

        self.assertEqual(manager.get_installed_extensions(),
                         [HooksTestExtension])
        self.assertEqual(manager.get_enabled_extensions(), [extension])
        self.assertIs(HooksTestExtension.instance, extension)
        self.assertEqual(len(URLHook.hooks), 1)
        self.assertEqual(manager._load_errors, {})

        @self.spy_for(manager._install_admin_urls)
        def _install_admin_urls(*args, **kwargs):
            raise Exception('admin URLs failed!')

        with self.scanned_extensions([HooksTestExtension]):
            self.extension_mgr.load(full_reload=True)

        self.assertEqual(manager.get_installed_extensions(),
                         [HooksTestExtension])
        self.assertEqual(manager.get_enabled_extensions(), [])
        self.assertIsNone(HooksTestExtension.instance)
        self.assertEqual(len(URLHook.hooks), 0)
        self.assertIn(extension.id, manager._load_errors)
        self.assertTrue(manager._load_errors[extension.id].startswith(
            'admin URLs failed!\n\nTraceback'))

    def test_sync_database_with_no_settings_version(self):
        """Testing ExtensionManager synchronizes database when no version
        found in settings (new install)
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(MyTestExtension, enable=False)
        extension = MyTestExtension(extension_mgr)
        extension.registration.installed = True
        extension.registration.save()

        self.assertNotIn(extension_mgr.VERSION_SETTINGS_KEY,
                         extension.settings)

        self.spy_on(extension_mgr._sync_database, call_original=False)

        extension = extension_mgr.enable_extension(MyTestExtension.id)

        self.assertTrue(extension.registration.installed)
        self.assertSpyCalled(extension_mgr._sync_database)
        self.assertIn(extension_mgr.VERSION_SETTINGS_KEY, extension.settings)
        self.assertEqual(
            extension.settings[extension_mgr.VERSION_SETTINGS_KEY],
            extension.info.version)

    def test_sync_database_with_old_settings_version(self):
        """Testing ExtensionManager synchronizes database when old version
        found in settings (upgrade)
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(MyTestExtension, enable=False)
        extension = MyTestExtension(extension_mgr)
        extension.registration.installed = True
        extension.registration.save()

        self.assertNotIn(extension_mgr.VERSION_SETTINGS_KEY,
                         extension.settings)
        extension.settings.set(extension_mgr.VERSION_SETTINGS_KEY, '0.5')
        extension.settings.save()

        self.spy_on(extension_mgr._sync_database, call_original=False)

        extension = extension_mgr.enable_extension(MyTestExtension.id)

        self.assertTrue(extension.registration.installed)
        self.assertSpyCalled(extension_mgr._sync_database)
        self.assertIn(extension_mgr.VERSION_SETTINGS_KEY, extension.settings)
        self.assertEqual(
            extension.settings[extension_mgr.VERSION_SETTINGS_KEY],
            extension.info.version)

    def test_sync_database_with_current_settings_version(self):
        """Testing ExtensionManager doesn't synchronize database when current
        version found in settings
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(MyTestExtension, enable=False)
        extension = MyTestExtension(extension_mgr)
        extension.registration.installed = True
        extension.registration.save()

        self.assertNotIn(extension_mgr.VERSION_SETTINGS_KEY,
                         extension.settings)
        extension.settings.set(extension_mgr.VERSION_SETTINGS_KEY,
                               extension.info.version)
        extension.settings.save()

        self.spy_on(extension_mgr._sync_database, call_original=False)

        extension = extension_mgr.enable_extension(MyTestExtension.id)

        self.assertTrue(extension.registration.installed)
        self.assertSpyNotCalled(extension_mgr._sync_database)
        self.assertIn(extension_mgr.VERSION_SETTINGS_KEY, extension.settings)
        self.assertEqual(
            extension.settings[extension_mgr.VERSION_SETTINGS_KEY],
            extension.info.version)

    def test_sync_database_with_newer_settings_version(self):
        """Testing ExtensionManager doesn't synchronize database when newer
        version found in settings (downgrade)
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(MyTestExtension, enable=False)
        extension = MyTestExtension(extension_mgr)
        extension.registration.installed = True
        extension.registration.save()

        self.assertNotIn(extension_mgr.VERSION_SETTINGS_KEY,
                         extension.settings)
        extension.settings.set(extension_mgr.VERSION_SETTINGS_KEY, '100.0')
        extension.settings.save()

        self.spy_on(extension_mgr._sync_database, call_original=False)
        self.spy_on(manager_logger.warning)

        extension = extension_mgr.enable_extension(MyTestExtension.id)

        self.assertTrue(extension.registration.installed)
        self.assertSpyNotCalled(extension_mgr._sync_database)
        self.assertIn(extension_mgr.VERSION_SETTINGS_KEY, extension.settings)
        self.assertEqual(
            extension.settings[extension_mgr.VERSION_SETTINGS_KEY],
            '100.0')
        self.assertIn('is older than the version recorded',
                      manager_logger.warning.spy.calls[-1].args[0])

    def test_install_extension_media_with_no_version_file(self):
        """Testing ExtensionManager installs media when no version file exists
        in the media directory (new install or upgrade from pre-version file)
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(MyTestExtension, enable=False)
        extension = MyTestExtension(extension_mgr)
        extension.registration.installed = True
        extension.registration.enabled = True
        extension.registration.save()
        MyTestExtension.instance = extension

        self._rebuild_media_dirs(extension)

        extension_mgr.should_install_static_media = True

        self.spy_on(extension_mgr.install_extension_media)
        self.spy_on(extension_mgr._install_extension_media_internal,
                    call_original=False)

        # Fake the current version being installed.
        version_filename = os.path.join(extension.info.installed_static_path,
                                        '.version')
        self.assertFalse(os.path.exists(version_filename))

        # Enable the extension. We shouldn't install anything.
        extension = extension_mgr.enable_extension(MyTestExtension.id)
        self.assertTrue(extension.registration.installed)

        self.assertSpyCallCount(extension_mgr.install_extension_media, 1)
        self.assertSpyCallCount(
            extension_mgr._install_extension_media_internal, 1)

        with open(version_filename, 'r') as fp:
            self.assertEqual(fp.read().strip(), extension.info.version)

    def test_install_extension_media_with_outdated_version(self):
        """Testing ExtensionManager installs media when version file is
        older than the current version (upgrade)
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(MyTestExtension, enable=False)
        extension = MyTestExtension(extension_mgr)
        extension.registration.installed = True
        extension.registration.enabled = True
        extension.registration.save()
        MyTestExtension.instance = extension

        self._rebuild_media_dirs(extension)

        extension_mgr.should_install_static_media = True

        self.spy_on(extension_mgr.install_extension_media)
        self.spy_on(extension_mgr._install_extension_media_internal,
                    call_original=False)

        # Fake the current version being installed.
        version_filename = os.path.join(extension.info.installed_static_path,
                                        '.version')

        with open(version_filename, 'w') as fp:
            fp.write('0.5\n')

        # Enable the extension. We shouldn't install anything.
        extension = extension_mgr.enable_extension(MyTestExtension.id)
        self.assertTrue(extension.registration.installed)

        self.assertSpyCallCount(extension_mgr.install_extension_media, 1)
        self.assertSpyCallCount(
            extension_mgr._install_extension_media_internal, 1)

        with open(version_filename, 'r') as fp:
            self.assertEqual(fp.read().strip(), extension.info.version)

    def test_install_extension_media_with_newer_version(self):
        """Testing ExtensionManager installs media when version file is
        newer than the current version (downgrade)
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(MyTestExtension, enable=False)
        extension = MyTestExtension(extension_mgr)
        extension.registration.installed = True
        extension.registration.enabled = True
        extension.registration.save()
        MyTestExtension.instance = extension

        self._rebuild_media_dirs(extension)

        extension_mgr.should_install_static_media = True

        self.spy_on(extension_mgr.install_extension_media)
        self.spy_on(extension_mgr._install_extension_media_internal,
                    call_original=False)

        # Fake the current version being installed.
        version_filename = os.path.join(extension.info.installed_static_path,
                                        '.version')

        with open(version_filename, 'w') as fp:
            fp.write('100.0\n')

        # Enable the extension. We shouldn't install anything.
        extension = extension_mgr.enable_extension(MyTestExtension.id)
        self.assertTrue(extension.registration.installed)

        self.assertSpyCallCount(extension_mgr.install_extension_media, 1)
        self.assertSpyCallCount(
            extension_mgr._install_extension_media_internal, 1)

        with open(version_filename, 'r') as fp:
            self.assertEqual(fp.read().strip(), extension.info.version)

    def test_install_extension_media_with_permission_error(self):
        """Testing ExtensionManager media installation with permission error
        when copying media files
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(HooksTestExtension, enable=False)
        extension = HooksTestExtension(extension_mgr)
        extension.shutdown()

        self.assertEqual(len(URLHook.hooks), 0)

        self._rebuild_media_dirs(extension)

        extension_mgr.should_install_static_media = True

        @self.spy_for(shutil.copytree)
        def _copytree(*args, **kwargs):
            raise shutil.Error()

        self.spy_on(extension_mgr.install_extension_media)
        self.spy_on(extension.info.has_resource,
                    op=kgb.SpyOpReturn(True))

        message = (
            'Unable to install static media files for this extension. '
            'The extension will not work correctly without them. Please '
            'make sure that "%s", its contents, and its parent directory '
            'are owned by the web server.'
        ) % os.path.join(
            settings.STATIC_ROOT, 'ext',
            'djblets.extensions.tests.test_extension_manager.'
            'HooksTestExtension'
        )

        with self.assertRaisesMessage(EnablingExtensionError, message):
            extension_mgr.enable_extension(HooksTestExtension.id)

        self.assertSpyCallCount(extension_mgr.install_extension_media, 1)
        self.assertFalse(os.path.exists(
            os.path.join(extension.info.installed_static_path, '.version')))
        self.assertEqual(extension_mgr.get_installed_extensions(),
                         [HooksTestExtension])
        self.assertEqual(extension_mgr.get_enabled_extensions(), [])
        self.assertIsNone(HooksTestExtension.instance)
        self.assertEqual(len(URLHook.hooks), 0)

    def test_install_extension_media_with_lock_error(self):
        """Testing ExtensionManager media installation with error when locking
        static directory
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(HooksTestExtension, enable=False)
        extension = HooksTestExtension(extension_mgr)
        extension.shutdown()

        self._rebuild_media_dirs(extension)

        extension_mgr.should_install_static_media = True
        extension_mgr._MEDIA_LOCK_SLEEP_TIME_SECS = 0

        @self.spy_for(extension_mgr._open_lock_file)
        def _open_lock_file(*args, **kwargs):
            raise IOError(errno.EAGAIN, 'Try again')

        self.spy_on(extension_mgr.install_extension_media)
        self.spy_on(extension_mgr._install_extension_media_internal,
                    call_original=False)

        message = (
            'Unable to install static media files for this extension. There '
            'have been 10 attempts to install the media files. Please make '
            'sure that "%s", its contents, its parent directory, and "%s" are '
            'writable by the web server.'
        ) % (
            os.path.join(
                settings.STATIC_ROOT, 'ext',
                'djblets.extensions.tests.test_extension_manager'
                '.HooksTestExtension'
            ),
            tempfile.gettempdir()
        )

        with self.assertRaisesMessage(EnablingExtensionError, message):
            extension_mgr.enable_extension(HooksTestExtension.id)

        self.assertSpyCallCount(extension_mgr.install_extension_media, 1)
        self.assertSpyNotCalled(
            extension_mgr._install_extension_media_internal)

        self.assertFalse(os.path.exists(
            os.path.join(extension.info.installed_static_path, '.version')))

        self.assertEqual(extension_mgr.get_installed_extensions(),
                         [HooksTestExtension])
        self.assertEqual(extension_mgr.get_enabled_extensions(), [])
        self.assertIsNone(HooksTestExtension.instance)
        self.assertEqual(len(URLHook.hooks), 0)

    def test_install_media_concurrent_threads(self):
        """Testing ExtensionManager installs media safely with multiple
        threads
        """
        extension_mgr = self.extension_mgr

        # Manually mark the extension as installed and enabled without going
        # through the enable process, since we don't want to trigger any
        # static media saving yet.
        self.setup_extension(MyTestExtension, enable=False)
        extension = MyTestExtension(extension_mgr)
        extension.registration.installed = True
        extension.registration.enabled = True
        extension.registration.save()
        MyTestExtension.instance = extension

        self._rebuild_media_dirs(extension)

        extension_mgr.should_install_static_media = True

        self.assertEqual(len(extension_mgr.get_installed_extensions()), 1)

        self.spy_on(extension_mgr.install_extension_media)
        self.spy_on(extension_mgr._install_extension_media_internal,
                    call_original=False)

        # Simulate numerous simultaneous attempts at installing static media.
        self._run_thread_test(
            lambda: extension_mgr.install_extension_media(extension.__class__))

        self.assertEqual(
            len(extension_mgr.install_extension_media.calls), 2)
        self.assertEqual(
            len(extension_mgr._install_extension_media_internal.calls), 1)
        self.assertEqual(self.exceptions, [])

        version_filename = os.path.join(extension.info.installed_static_path,
                                        '.version')

        with open(version_filename, 'r') as fp:
            self.assertEqual(fp.read().strip(), extension.info.version)

    def test_install_extension_media_with_current_version(self):
        """Testing ExtensionManager doesn't install media when version file
        matches current version
        """
        extension_mgr = self.extension_mgr

        self.setup_extension(MyTestExtension, enable=False)
        extension = MyTestExtension(extension_mgr)
        MyTestExtension.registration.installed = True
        MyTestExtension.registration.enabled = True
        MyTestExtension.registration.save()
        MyTestExtension.instance = extension

        self._rebuild_media_dirs(extension)

        extension_mgr.should_install_static_media = True

        self.spy_on(extension_mgr.install_extension_media)
        self.spy_on(extension_mgr._install_extension_media_internal,
                    call_original=False)

        # Fake the current version being installed.
        version_filename = os.path.join(extension.info.installed_static_path,
                                        '.version')

        with open(version_filename, 'w') as fp:
            fp.write('%s\n' % extension.info.version)

        # Enable the extension. We shouldn't install anything.
        extension = extension_mgr.enable_extension(MyTestExtension.id)
        self.assertTrue(extension.registration.installed)

        self.assertEqual(len(extension_mgr.install_extension_media.calls), 1)
        self.assertEqual(
            len(extension_mgr._install_extension_media_internal.calls), 0)

        with open(version_filename, 'r') as fp:
            self.assertEqual(fp.read().strip(), extension.info.version)

    def test_enable_extension_registers_static_bundles(self):
        """Testing ExtensionManager.enable_extension registers static bundles
        """
        class MyTestExtensionWithBundles(Extension):
            css_bundles = {
                'default': {
                    'source_filenames': ['test.css'],
                }
            }

            js_bundles = {
                'default': {
                    'source_filenames': ['test.js'],
                }
            }

        pipeline_settings.STYLESHEETS = {}
        pipeline_settings.JAVASCRIPT = {}

        extension = self.setup_extension(MyTestExtensionWithBundles)

        self.assertEqual(len(pipeline_settings.STYLESHEETS), 1)
        self.assertEqual(len(pipeline_settings.JAVASCRIPT), 1)

        key = '%s-default' % extension.id
        self.assertIn(key, pipeline_settings.STYLESHEETS)
        self.assertIn(key, pipeline_settings.JAVASCRIPT)

        css_bundle = pipeline_settings.STYLESHEETS[key]
        js_bundle = pipeline_settings.JAVASCRIPT[key]

        self.assertIn('source_filenames', css_bundle)
        self.assertEqual(css_bundle['source_filenames'],
                         ['ext/%s/test.css' % extension.id])

        self.assertIn('output_filename', css_bundle)
        self.assertEqual(css_bundle['output_filename'],
                         'ext/%s/css/default.min.css' % extension.id)

        self.assertIn('source_filenames', js_bundle)
        self.assertEqual(js_bundle['source_filenames'],
                         ['ext/%s/test.js' % extension.id])

        self.assertIn('output_filename', js_bundle)
        self.assertEqual(js_bundle['output_filename'],
                         'ext/%s/js/default.min.js' % extension.id)

    def test_enable_extension_registers_context_processors(self):
        """Testing ExtensionManager.enable_extension registers template
        context processors
        """
        class MyTestExtensionWithContextProcessors(Extension):
            context_processors = ['my_custom_processor']

        # Back up the list, so we can replace it later.
        if hasattr(settings, 'TEMPLATES'):
            orig_context_processors_list = \
                list(settings.TEMPLATES[0]['OPTIONS']['context_processors'])
        else:
            orig_context_processors_list = \
                list(settings.TEMPLATE_CONTEXT_PROCESSORS)

        # Sanity-check that the context processor didn't wind up in here.
        self.assertNotIn('my_custom_processor', orig_context_processors_list)

        try:
            extension = self.setup_extension(
                MyTestExtensionWithContextProcessors)

            # We have to re-fetch these lists now, since they may have
            # been normalized to lists.
            if hasattr(settings, 'TEMPLATES'):
                context_processors_list = \
                    settings.TEMPLATES[0]['OPTIONS']['context_processors']
            else:
                context_processors_list = \
                    settings.TEMPLATE_CONTEXT_PROCESSORS

            # This should have been added, since the extension was enabled.
            self.assertIn('my_custom_processor', context_processors_list)

            # Shutting down the extension should remove the context
            # processor.
            self.extension_mgr.disable_extension(extension.id)
            self.assertNotIn('my_custom_processor',
                             context_processors_list)
        finally:
            if hasattr(settings, 'TEMPLATES'):
                settings.TEMPLATES[0]['OPTIONS']['context_processors'] = \
                    orig_context_processors_list
            else:
                settings.TEMPLATE_CONTEXT_PROCESSORS = \
                    orig_context_processors_list

    def test_enable_extension_registers_template_tags(self):
        """Testing ExtensionManager.enable_extension registers template tags"""
        class MyTestExtensionWithModule(Extension):
            __module__ = ('djblets.extensions.tests.apps'
                          '.templatetag_tests.__init__')

        templatetags_module = \
            'djblets.extensions.tests.apps.templatetag_tests.templatetags'

        def _check_state(enabled):
            if enabled:
                self.assertEqual(
                    Template(template_str).render(Context({})),
                    'Hello, world!')
            else:
                with self.assertRaisesRegex(TemplateSyntaxError,
                                            'is not a (valid|registered) tag '
                                            'library'):
                    Template(template_str).render(Context({}))

        template_str = (
            '{% load templatetag_tests %}'
            '{% my_extension_template_tag %}'
        )

        # Sanity-check that the template tag module isn't registered.
        _check_state(enabled=False)

        # Enabling the extension should register the template tags module and
        # clear the cache.
        extension = self.setup_extension(MyTestExtensionWithModule)

        _check_state(enabled=True)

        # Shutting down the extension should remove the template tags module
        # and clear the cache.
        self.extension_mgr.disable_extension(extension.id)

        _check_state(enabled=False)

        # Other libraries should still work.
        Template('{% load djblets_js djblets_extensions %}').render(
            Context({}))

    def test_enable_extension_evolve_with_models(self):
        """Testing ExtensionManager.enable_extension synchronizes database
        models
        """
        self.spy_on(Evolver.evolve,
                    owner=Evolver)

        class MyTestExtensionWithApps(Extension):
            apps = [
                'djblets.extensions.tests.apps.model_tests',
            ]

        extension = self.setup_extension(MyTestExtensionWithApps)

        self.assertTrue(Evolver.evolve.called)

        from djblets.extensions.tests.apps.model_tests.models import \
            MyTestExtensionModel

        # We should be able to create entries and query them.
        MyTestExtensionModel.objects.create(test_field='test')
        self.assertEqual(MyTestExtensionModel.objects.count(), 1)

        # Re-enabling shouldn't break anything.
        self.extension_mgr.disable_extension(extension.id)
        self.extension_mgr.enable_extension(extension.id)
        self.assertEqual(MyTestExtensionModel.objects.count(), 1)

    def test_enable_extension_evolve_without_models(self):
        """Testing ExtensionManager.enable_extension does not synchronize
        database models if extension does not have any
        """
        self.spy_on(Evolver.evolve,
                    owner=Evolver)

        class MyTestExtensionWithApps(Extension):
            apps = [
                'djblets.extensions.tests.apps',
            ]

        self.setup_extension(MyTestExtensionWithApps)

        self.assertFalse(Evolver.evolve.called)

    def test_enable_extension_evolve_with_pending_evolutions(self):
        """Testing ExtensionManager.enable_extension evolves database models
        when pending evolutions found
        """
        from django_evolution.models import Version
        from django_evolution.signature import AppSignature, ModelSignature

        self.spy_on(Evolver.evolve,
                    owner=Evolver)

        class MyTestExtensionWithApps(Extension):
            apps = [
                'djblets.extensions.tests.apps.evolve_tests',
            ]

        # We need to set some initial state in the database for the model and
        # for the evolution history.
        connection.cursor().execute(
            'CREATE TABLE evolve_tests_testevolveextensionmodel ('
            '    id INTEGER PRIMARY KEY AUTOINCREMENT,'
            '    test_field VARCHAR(16) NOT NULL'
            ')')

        from djblets.extensions.tests.apps.evolve_tests.models import \
            TestEvolveExtensionModel

        latest_version = Version.objects.current_version()

        model_sig = ModelSignature.from_model(TestEvolveExtensionModel)
        model_sig.remove_field_sig('new_field')

        app_sig = AppSignature(app_id='evolve_tests')
        app_sig.add_model_sig(model_sig)

        latest_version.signature.add_app_sig(app_sig)
        latest_version.save()

        # We can now enable the extension, which will perform an evolution.
        extension = self.setup_extension(MyTestExtensionWithApps)

        self.assertTrue(Evolver.evolve.called)

        # We should be able to create entries and query them.
        TestEvolveExtensionModel.objects.create(test_field='test')
        self.assertEqual(TestEvolveExtensionModel.objects.count(), 1)

        # We're now going to shut down and re-enable, but with a different
        # version of the model. This should trigger an evolution sequence.
        self.extension_mgr.disable_extension(extension.id)

        self.extension_mgr.enable_extension(extension.id)

        TestEvolveExtensionModel.objects.create(test_field='test',
                                                new_field=100)
        self.assertEqual(TestEvolveExtensionModel.objects.count(), 2)

        obj = TestEvolveExtensionModel.objects.get(pk=2)
        self.assertEqual(obj.new_field, 100)

    def test_enable_extension_evolve_with_applied_evolutions(self):
        """Testing ExtensionManager.enable_extension evolves database models
        when all evolutions are already applied
        """
        from django_evolution.models import Evolution, Version
        from django_evolution.signature import AppSignature, ModelSignature

        self.spy_on(Evolver.evolve,
                    owner=Evolver)

        class MyTestExtensionWithApps(Extension):
            apps = [
                'djblets.extensions.tests.apps.evolve_tests',
            ]

        # We need to set some initial state in the database for the model and
        # for the evolution history.
        connection.cursor().execute(
            'CREATE TABLE evolve_tests_testevolveextensionmodel ('
            '    id INTEGER PRIMARY KEY AUTOINCREMENT,'
            '    test_field VARCHAR(16) NOT NULL'
            ')')

        from djblets.extensions.tests.apps.model_tests.models import \
            MyTestExtensionModel

        latest_version = Version.objects.current_version()
        signature = latest_version.signature.clone()

        model_sig = ModelSignature.from_model(MyTestExtensionModel)
        model_sig.model_name = 'TestEvolveExtensionModel'
        model_sig.table_name = 'evolve_tests_testevolveextensionmodel'

        app_sig = AppSignature(app_id='evolve_tests')
        app_sig.add_model_sig(model_sig)
        signature.add_app_sig(app_sig)

        version = Version.objects.create(signature=signature)
        Evolution.objects.create(version=version,
                                 app_label='evolve_tests',
                                 label='add_new_field')

        # We can now enable the extension, which will perform an evolution.
        self.setup_extension(MyTestExtensionWithApps)

        self.assertFalse(Evolver.evolve.called)

    def test_disable_unregisters_static_bundles(self):
        """Testing ExtensionManager unregisters static bundles when disabling
        extension
        """
        class MyTestExtensionWithBundles(Extension):
            css_bundles = {
                'default': {
                    'source_filenames': ['test.css'],
                }
            }

            js_bundles = {
                'default': {
                    'source_filenames': ['test.js'],
                }
            }

        pipeline_settings.STYLESHEETS = {}
        pipeline_settings.JAVASCRIPT = {}

        extension = self.setup_extension(MyTestExtensionWithBundles)

        self.assertEqual(len(pipeline_settings.STYLESHEETS), 1)
        self.assertEqual(len(pipeline_settings.JAVASCRIPT), 1)

        self.extension_mgr.disable_extension(extension.id)

        self.assertEqual(len(pipeline_settings.STYLESHEETS), 0)
        self.assertEqual(len(pipeline_settings.JAVASCRIPT), 0)

    def test_extension_list_sync(self):
        """Testing ExtensionManager extension list synchronization
        cross-process
        """
        key = 'extension-list-sync'

        manager1 = ExtensionManager(key)
        manager2 = ExtensionManager(key)

        with self.scanned_extensions([MyTestExtension],
                                     extension_mgr=manager1):
            manager1.load()

        with self.scanned_extensions([MyTestExtension],
                                     extension_mgr=manager2):
            manager2.load()

        self.assertEqual(len(manager1.get_installed_extensions()), 1)
        self.assertEqual(len(manager2.get_installed_extensions()), 1)
        self.assertEqual(len(manager1.get_enabled_extensions()), 0)
        self.assertEqual(len(manager2.get_enabled_extensions()), 0)

        manager1.enable_extension(MyTestExtension.id)
        self.assertEqual(len(manager1.get_enabled_extensions()), 1)
        self.assertEqual(len(manager2.get_enabled_extensions()), 0)

        self.assertFalse(manager1.is_expired())
        self.assertTrue(manager2.is_expired())

        with self.scanned_extensions([MyTestExtension],
                                     extension_mgr=manager2):
            manager2.load(full_reload=True)

        self.assertEqual(len(manager1.get_enabled_extensions()), 1)
        self.assertEqual(len(manager2.get_enabled_extensions()), 1)
        self.assertFalse(manager1.is_expired())
        self.assertFalse(manager2.is_expired())

    def test_extension_settings_sync(self):
        """Testing ExtensionManager extension settings synchronization
        cross-process
        """
        key = 'extension-settings-sync'
        setting_key = 'foo'
        setting_val = 'abc123'

        manager1 = ExtensionManager(key)
        manager2 = ExtensionManager(key)

        with self.scanned_extensions([MyTestExtension],
                                     extension_mgr=manager1):
            manager1.load()

        extension1 = manager1.enable_extension(MyTestExtension.id)

        with self.scanned_extensions([MyTestExtension],
                                     extension_mgr=manager2):
            manager2.load()

        self.assertFalse(manager1.is_expired())
        self.assertFalse(manager2.is_expired())

        extension2 = manager2.get_enabled_extension(MyTestExtension.id)
        self.assertNotEqual(extension2, None)

        self.assertNotIn(setting_key, extension1.settings)
        self.assertNotIn(setting_key, extension2.settings)
        extension1.settings[setting_key] = setting_val
        extension1.settings.save()

        self.assertNotIn(setting_key, extension2.settings)

        self.assertFalse(manager1.is_expired())
        self.assertTrue(manager2.is_expired())

        with self.scanned_extensions([MyTestExtension],
                                     extension_mgr=manager2):
            manager2.load(full_reload=True)

        extension2 = manager2.get_enabled_extension(MyTestExtension.id)

        self.assertFalse(manager1.is_expired())
        self.assertFalse(manager2.is_expired())
        self.assertIn(setting_key, extension1.settings)
        self.assertIn(setting_key, extension2.settings)
        self.assertEqual(extension1.settings[setting_key], setting_val)
        self.assertEqual(extension2.settings[setting_key], setting_val)

    def test_load_blocks_sync_gen(self):
        """Testing ExtensionManager.load blocks bumping sync generation
        number
        """
        key = 'check-expired-test'
        manager1 = ExtensionManager(key)
        manager2 = ExtensionManager(key)

        with self.scanned_extensions([MyTestExtension],
                                     extension_mgr=manager1):
            manager1.load()

        manager1.enable_extension(MyTestExtension.id)

        with self.scanned_extensions([MyTestExtension],
                                     extension_mgr=manager2):
            manager2.load()

        self.assertEqual(manager1._gen_sync.sync_gen,
                         manager2._gen_sync.sync_gen)

        # Trigger a save whenever the extension initializes.
        MyTestExtension.initialize = lambda ext: ext.settings.save()

        # Bump the generation number.
        extension = manager2.get_enabled_extension(MyTestExtension.id)
        extension.settings.save()
        self.assertNotEqual(manager1._gen_sync.sync_gen,
                            manager2._gen_sync.sync_gen)

        # Loading now should retain the new sync generation number, instead
        # of bumping it.
        with self.scanned_extensions([MyTestExtension],
                                     extension_mgr=manager1):
            manager1.load(full_reload=True)

        self.assertEqual(manager1._gen_sync.sync_gen,
                         manager2._gen_sync.sync_gen)

    def _rebuild_media_dirs(self, extension):
        """Rebuild the static media directories for an extension.

        This will erase the existing directories, if they exist, and then
        create a new directory for the tests.

        Args:
            extension (djblets.extensions.extension.Extension):
                The extension being tested.
        """
        path = extension.info.installed_static_path

        self.assertTrue(os.path.isabs(path))

        if os.path.exists(path):
            shutil.rmtree(path)

        os.makedirs(extension.info.installed_static_path, 0o755)

    def _run_thread_test(self, main_func):
        def _thread_main(main_connection, main_func, sleep_time):
            # Insert the connection from the main thread, so that we can
            # perform lookups. We never write.
            from django.db import connections

            connections['default'] = main_connection

            time.sleep(sleep_time)
            main_func()

        # Store the main connection. We're going to let the threads share it.
        # This trick courtesy of the Django unit tests
        # (django/tests/backends/tests.py).
        from django.db import connections

        main_connection = connections['default']

        try:
            main_connection.inc_thread_sharing()

            self.exceptions = []

            t1 = threading.Thread(target=_thread_main,
                                  args=[main_connection, main_func, 0.2])
            t2 = threading.Thread(target=_thread_main,
                                  args=[main_connection, main_func, 0.1])
            t1.start()
            t2.start()
            t1.join()
            t2.join()
        finally:
            main_connection.dec_thread_sharing()

    def _sleep_and_call(self, manager, orig_func, *args, **kwargs):
        # This works well enough to throw a monkey wrench into things.
        # One thread will be slightly ahead of the other.
        time.sleep(0.2)

        try:
            orig_func(*args, **kwargs)
        except Exception as e:
            logging.error('%s\n', e, exc_info=1)
            self.exceptions.append(e)

    def _spy_sleep_and_call(self, func):
        def _call(manager, *args, **kwargs):
            self._sleep_and_call(manager, orig_func, *args, **kwargs)

        orig_func = func

        self.spy_on(func, call_fake=_call)



"""Views for working with site configurations."""

from __future__ import annotations

from typing import Dict, Type

from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.views.decorators.csrf import csrf_protect

from djblets.siteconfig.forms import SiteSettingsForm
from djblets.siteconfig.models import SiteConfiguration


@csrf_protect
@staff_member_required
def site_settings(
    request: HttpRequest,
    form_class: Type[SiteSettingsForm] = SiteSettingsForm,
    template_name: str = 'siteconfig/settings.html',
    extra_context: Dict = {},
) -> HttpResponse:
    """A view for customizing the site configuration's settings.

    This will make use of the provided form class and template, handling
    loading and saving any site configuration settings.

    Args:
        request (django.http.HttpRequest):
            The HTTP request for the view.

        form_class (type, optional):
            The form class to use for editing the settings. This should be
            either :py:class:`~djblets.siteconfig.forms.SiteSettingsForm` or
            a subclass of it.

        template_name (str, optional):
            The template to render for the view.

        extra_context (dict, optional):
            Extra context variables to pass to the template.

    Returns:
        django.http.HttpResponse:
        The response containing the page.
    """
    siteconfig = SiteConfiguration.objects.get_current()

    if request.method == 'POST':
        form = form_class(siteconfig, request.POST, request.FILES,
                          request=request)

        if form.is_valid():
            form.save()
            return HttpResponseRedirect('.?saved=1')
    else:
        form = form_class(siteconfig, request=request)

    context = {
        'form': form,
        'opts': siteconfig._meta,
        'saved': request.GET.get('saved', 0)
    }
    context.update(extra_context)

    return render(request, template_name, context)



from django import template
from django.utils.html import mark_safe


register = template.Library()


@register.simple_tag
def recaptcha_js():
    """Render the reCAPTCHA JavaScript tag.

    Returns:
        django.utils.safestring.SafeText:
        The rendered tag.
    """
    return mark_safe('<script src="https://www.google.com/recaptcha/api.js">'
                     '</script>')


@register.simple_tag
def recaptcha_form_field(form):
    """Return the reCAPTCHA field from the specified form.

    This can be used to render the reCAPTCHA widget.

    Args:
        form (django.forms.forms.Form):
            The form that is being rendered.

    Returns:
        django.forms.boundfield.BoundField:
        The bound reCAPTCHA field. This will render as its widget in a
        template.
    """
    return form['g-recaptcha-response']



"""Utilities for working with dates."""

import calendar
from datetime import datetime

from django.db.models import DateField
from django.utils.timezone import utc


def http_date(timestamp):
    """
    A wrapper around Django's http_date that accepts DateFields and
    datetime objects directly.
    """
    from django.utils.http import http_date

    if isinstance(timestamp, (DateField, datetime)):
        return http_date(calendar.timegm(timestamp.timetuple()))
    elif isinstance(timestamp, str):
        return timestamp
    else:
        return http_date(timestamp)


def get_latest_timestamp(timestamps):
    """
    Returns the latest timestamp in a list of timestamps.
    """
    latest = None

    for timestamp in timestamps:
        if latest is None or timestamp > latest:
            latest = timestamp

    return latest


def get_tz_aware_utcnow():
    """Returns a UTC aware datetime object"""
    return datetime.utcnow().replace(tzinfo=utc)



"""Tests for authentication-related views."""

from django.core.cache import cache
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.test.client import Client
from kgb import SpyAgency

from djblets.auth.forms import RegistrationForm
from djblets.testing.testcases import TestCase


class ViewsTests(SpyAgency, TestCase, Client):
    """Tests for djblets.auth.views"""

    def tearDown(self):
        super(ViewsTests, self).tearDown()
        cache.clear()

    def test_register_without_post_method(self):
        """Testing register view when request is not a post method"""
        self.spy_on(render)

        response = self.client.get('/auth/register/')

        self.assertIsNotNone(response.context)
        self.assertIsInstance(response.context['form'], RegistrationForm)
        self.assertTrue(render.called)

    def test_register_with_valid_form(self):
        """Testing register view when a valid form is received"""
        request_headers = {
            'username': 'user123',
            'password1': 'password1',
            'password2': 'password1',
            'email': 'test@mailinator.com',
        }
        # Make a get request first to ensure we have a csrf token
        # generated. By doing so, we can ensure our post request and form
        # are valid.
        self.client.get('/auth/register/')

        self.spy_on(render)

        response = self.client.post('/auth/register/', request_headers)

        self.assertIsInstance(response, HttpResponseRedirect)
        self.assertIsNone(response.context)
        self.assertFalse(render.called)

    def test_register_with_invalid_form(self):
        """Testing register view when an invalid form is received"""
        request_headers = {
            'username': 'user123',
            'password1': 'password1',
            'password2': 'password1',
            'email': 'invalid email',
        }
        self.client.get('/auth/register/')

        self.spy_on(render)

        response = self.client.post('/auth/register/', request_headers)

        self.assertIsNotNone(response.context)
        self.assertIsInstance(response.context['form'], RegistrationForm)
        self.assertTrue(render.called)



"""Form-related template tags."""

from django import forms, template
from django.utils.html import conditional_escape, format_html


register = template.Library()


@register.simple_tag
def label_tag(field):
    """Render the tag for a field's label.

    This generates labels similar to the administration UI's own field labels,
    providing styling for required fields and checkboxes.

    If the label is explicitly set to an empty string on the field, no label
    will be rendered, matching Django's default behavior.

    Args:
        field (django.forms.BoundField):
            The bound field on the form to render the label for.

    Returns:
        django.utils.safestring.SafeText:
        The resulting HTML for the label.

    Example:
        .. code-block:: html+django

           {% label_for form.my_field %}
    """
    if not field.label:
        return ''

    is_checkbox = is_field_checkbox(field)

    # Build the list of CSS classes to apply to the label.
    classes = []

    if field.field.required:
        classes.append('required')

    if is_checkbox:
        classes.append('vCheckboxLabel')

    if classes:
        classes_html = format_html(' class="{0}"', ' '.join(classes))
    else:
        classes_html = ''

    # Build the text for the label.
    label = field.label

    if not is_checkbox:
        label += ':'

    # Render the result.
    return format_html('<label for="{0}"{1}>{2}</label>',
                       form_field_id(field),
                       classes_html,
                       label)


@register.filter
def form_field_id(field):
    """Render the ID of a field.

    This will derive the field's ID in the form and output it, for use in
    utility functions or custom HTML.

    Args:
        field (django.forms.BoundField):
            The bound field on the form.

    Returns:
        django.utils.safestring.SafeText:
        The resulting ID as safe HTML.

    Example:
        .. code-block:: html+django

           <span data-field-id="{{form.my_field|form_field_id}}"></span>
    """
    widget = field.field.widget
    field_id = widget.attrs.get('id') or field.auto_id or ''

    if field_id:
        field_id = widget.id_for_label(field_id)

    return conditional_escape(field_id)


@register.filter
def is_field_checkbox(field):
    """Return whether or not this field is effectively a checkbox field.

    A field is considered to be a checkbox field if its widget is a
    :py:class:`~django.forms.CheckboxInput`.

    Args:
        field (django.forms.BoundField):
            The bound field on the form.

    Returns:
        bool:
        ``True`` if this is a checkbox field. ``False`` if not.

    Example:
        .. code-block:: html+django

           {% if field|is_field_checkbox %}
           ...
           {% endif %}
    """
    return isinstance(field.field.widget, forms.CheckboxInput)


@register.filter
def is_checkbox_row(field):
    """Return whether the field's row is a checkbox-ish row.

    This will return ``True`` if rendering a checkbox, radio button, or
    multi-select checkbox.

    Args:
        field (django.forms.BoundField):
            The bound field on the form.

    Returns:
        bool:
        ``True`` if this is a checkbox-ish field. ``False`` if not.

    Example:
        .. code-block:: html+django

           {% if field|is_checkbox_row %}
           ...
           {% endif %}
    """
    return isinstance(field.field.widget, (forms.CheckboxInput,
                                           forms.RadioSelect,
                                           forms.CheckboxSelectMultiple))


@register.filter
def form_field_has_label_first(field):
    """Return whether a form label should be displayed before the widget.

    This helps when rendering labels and widgets in the correct order.
    Typically, non-checkbox widgets are preceded by a label, and this lets
    templates determine if that should be the case for a given field.

    Args:
        field (django.forms.BoundField):
            The bound field on the form.

    Returns:
        bool:
        ``True`` if the label should appear before the widget.
        ``False`` if the widget should appear before the label.

    Example:
        .. code-block:: html+django

           {% if field|form_field_has_label_first %}
           ...
           {% endif %}
    """
    return not is_field_checkbox(field)


@register.filter
def get_fieldsets(form):
    """Normalize and iterate over fieldsets in a form.

    This will loop through the fieldsets on a given form, converting either
    standard Django style or legay Djblets style fieldset data into a standard
    form and returning it to the template.

    Args:
        form (django.forms.Form):
            The form containing the fieldsets.

    Yields:
        tuple:
        A tuple of (fieldset_title, fieldset_info).

    Example:
        .. code-block:: html+django

           {% for fieldset_title, fieldset in form|get_fieldsets %}
           ...
           {% endfor %}
    """
    try:
        fieldsets = form.Meta.fieldsets
    except AttributeError:
        fieldsets = []

    for fieldset in fieldsets:
        if isinstance(fieldset, tuple):
            # This is a standard Django-style fieldset entry. It's a tuple
            # of (title, info).
            yield fieldset
        elif isinstance(fieldset, dict):
            # This is a legacy Djblets-style fieldset entry. It's a dictionary
            # that may contain the title as a "title" key.
            yield fieldset.get('title'), fieldset
        else:
            raise ValueError('Invalid fieldset value: %r' % fieldset)



"""Storage for static media built with django-pipeline.

Version Added:
    3.0
"""

from pipeline.storage import PipelineManifestStorage


class PipelineStorage(PipelineManifestStorage):
    """Storage for static media built with django-pipeline.

    This uses the ManifestStorage backend in order to add hashes to the
    filenames, but turns off strict mode so we don't need to list each piece of
    static media in the manifest file.

    Version Added:
        3.0
    """

    manifest_strict = False
    keep_intermediate_files = True



"""Base support and standard operators for condition choices."""

from django.utils.translation import gettext_lazy as _

from djblets.conditions.errors import (ConditionOperatorConflictError,
                                       ConditionOperatorNotFoundError)
from djblets.conditions.values import ConditionValueRegexField
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         ATTRIBUTE_REGISTERED, DEFAULT_ERRORS,
                                         NOT_REGISTERED, OrderedRegistry,
                                         UNREGISTER)


class BaseConditionOperator(object):
    """Base class for an operator for a condition choice.

    An operator forms an expression along with a parent
    :py:class:`~djblets.conditions.choices.BaseConditionChoice` and an optional
    value. This expression can be used by the condition to determine if a
    caller-provided value satisfies the condition.

    Choices will usually have more than one operator registered. Depending on
    the operator, there may or may not be a field for a value.

    Attributes:
        choice (djblets.conditions.choices.BaseConditionChoice):
            The choice owning the instance of this operation.
    """

    #: The ID of the operator.
    #:
    #: This must be unique within a
    #: :py:class:`~djblets.conditions.choices.BaseConditionChoice`.
    operator_id = None

    #: The displayed name for the operator.
    name = None

    @classmethod
    def with_overrides(cls, **attrs):
        """Dynamically create a subclass with overridden attributes.

        This makes it easy for a choice to make use of existing operators
        while using a custom name for display, or a custom value field,
        without having to create their own subclasses. It's meant only for
        simple changes.

        Args:
            **attrs (dict):
                Attributes to override on the operator.

        Returns:
            type:
            A new subclass with the overridden attributes.
        """
        return type(str('Custom%s') % cls.__name__, (cls,), attrs)

    def __init__(self, choice):
        """Initialize the operator.

        Args:
            choice (djblets.conditions.choices.BaseConditionChoice):
                The choice owning the instance of this operator.
        """
        self.choice = choice

    @property
    def value_field(self):
        """The field type used to prompt and render fields.

        By default, this will use the default one for the choice. The field
        can be disabled by setting this to ``None``, or a different field can
        be used by setting it to an instance of a
        :py:class:`~djblets.conditions.values.BaseConditionValueField`
        subclass or a function returning an instance.

        If it's a function, it must accept a ``**kwargs``, for future
        expansion.
        """
        default_value_field = self.choice.default_value_field

        if callable(default_value_field):
            return default_value_field()
        else:
            return default_value_field

    @property
    def has_custom_value_field(self):
        """Whether the operator has a custom value field."""
        return (self.__class__.value_field is not
                BaseConditionOperator.value_field)

    def matches(self, match_value, stored_value, **kwargs):
        """Return whether a value matches the operator and condition's value.

        This must be implemented by subclasses.

        Args:
            match_value (object):
                The caller's value to check against the state for this
                operator.

            condition_value (object):
                The value stored as part of the condition to check against.
                This will only be used if the operator has a value field
                associated.

            **kwargs (dict):
                Extra keyword arguments, for future expansion.

        Returns:
            bool:
            ``True`` if the lookup value fulfills the condition. ``False`` if
            it does not.

        Raises:
            TypeError:
                Either the lookup or condition value or types was not
                compatible with the expression.
        """
        raise NotImplementedError


class IsOneOfOperator(BaseConditionOperator):
    """An operator that matches against a set of possible values.

    This operator checks if the lookup value matches one of a set of possible
    values listed in the condition.

    This is equivalent to::

        if match_value in condition_value:
            ...
    """

    operator_id = 'one-of'
    name = _('Is one of')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value is one of a set of values.

        Args:
            match_value (object):
                The caller's value to check against the state for this
                operator.

            condition_value (list):
                A list of possible values that the lookup value must match.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value is present in the list of possible
            values.

        Raises:
            TypeError:
                Either the lookup or condition value was not compatible with
                the expression.
        """
        return match_value in condition_value


class IsNotOneOfOperator(BaseConditionOperator):
    """An operator that matches if not one of a set of possible values.

    This operator checks if the lookup value is not one of a set of possible
    values listed in the condition.

    This is equivalent to::

        if match_value not in condition_value:
            ...
    """

    operator_id = 'not-one-of'
    name = _('Is not one of')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value is not one of a set of values.

        Args:
            match_value (object):
                The caller's value to check against the state for this
                operator.

            condition_value (list):
                A list of possible values that the lookup value must not match.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value is not present in the list of possible
            values.

        Raises:
            TypeError:
                Either the lookup or condition value was not compatible with
                the expression.
        """
        return match_value not in condition_value


class AnyOperator(BaseConditionOperator):
    """An operator that matches for any non-empty/zero value.

    This operator checks if the lookup value provided is a non-empty value
    (a boolean, integer, string/list/dictionary containing a value, etc.).
    To determine if the value is empty, the operator checks ``len(value)``.
    If not 0, or if the value doesn't support a length check, it's assumed
    to have a value.

    This is equivalent to::

        if match_value in (0, False) or bool(match_value):
            ...

    The operator does not accept a user-provided condition value.

    This is the opposite of :py:class:`UnsetOperator`.
    """

    operator_id = 'any'
    name = _('Has a value')
    value_field = None

    def matches(self, match_value, **kwargs):
        """Return whether the lookup value is non-empty.

        Args:
            match_value (object):
                The caller's value to check against the state for this
                operator.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the value is evaluated as not empty.
            ``False`` if it evalutes as empty.
        """
        return match_value in (0, False) or bool(match_value)


class UnsetOperator(BaseConditionOperator):
    """An operator that matches for an unset/empty value.

    This operator checks if the lookup value provided is an empty value (such
    as empty string/list/dictionary or ``None``). This performs a simple
    ``not`` check against the value, but filters out values that evalute to
    ``False`` but are not considered unset.

    This is equivalent to::

        if match_value not in (0, False) and not match_value:
            ...

    The operator does not accept a user-provided condition value.

    This is the opposite of :py:class:`AnyOperator`.
    """

    operator_id = 'none'
    name = _('Is unset')
    value_field = None

    def matches(self, match_value, **kwargs):
        """Return whether the lookup value is empty.

        Args:
            match_value (object):
                The caller's value to check against the state for this
                operator.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the value is evaluated as empty.
            ``False`` if it evalutes as not empty.
        """
        return match_value not in (0, False) and not match_value


class IsOperator(BaseConditionOperator):
    """An operator that checks if one value is the same as another.

    This operator checks for equality, comparing the lookup value to the stored
    condition value.

    It's equivalent to::

        if match_value == condition_value:
            ...

    This is the opposite of :py:class:`IsNotOperator`.
    """

    operator_id = 'is'
    name = _('Is')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value equals a condition value.

        Args:
            match_value (object):
                The caller's value to check against the state for this
                operator.

            condition_value (object):
                The value to compare against.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value equals the condition value.
        """
        return match_value == condition_value


class IsNotOperator(BaseConditionOperator):
    """An operator that checks if one value is not the same as another.

    This operator checks for inequality, comparing the lookup value to the
    stored condition value.

    It's equivalent to::

        if match_value != condition_value:
            ...

    This is the opposite of :py:class:`IsOperator`.
    """

    operator_id = 'is-not'
    name = _('Is not')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value is not equal to a condition value.

        Args:
            match_value (object):
                The caller's value to check against the state for this
                operator.

            condition_value (object):
                The value to compare against.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value is not equal to the condition value.
        """
        return match_value != condition_value


class ContainsOperator(BaseConditionOperator):
    """An operator that checks if a lookup value contains a condition value.

    This operator checks if the provided lookup value contains the condition
    value within the value. It's useful for checking if a string or list
    contains some value.

    It's equivalent to::

        if condition_value in match_value:
            ...

    This is the opposite of :py:class:`DoesNotContainOperator`.
    """

    operator_id = 'contains'
    name = _('Contains')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value contains a condition value.

        Args:
            match_value (object):
                The caller's value to check for a condition value within.

            condition_value (object):
                The value to check within the lookup value.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value contains the condition value.

        Raises:
            TypeError:
                Either the lookup or condition value was not compatible with
                the expression.
        """
        return condition_value in match_value


class DoesNotContainOperator(BaseConditionOperator):
    """An operator that checks if a lookup value does not contain a value.

    This operator checks if the provided lookup value does not contain the
    condition value within the value. It's useful for checking if a string
    or list does not contain some value.

    It's equivalent to::

        if condition_value not in match_value:
            ...

    This is the opposite of :py:class:`ContainsOperator`.
    """

    operator_id = 'does-not-contain'
    name = _('Does not contain')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value does not contain a condition value.

        Args:
            match_value (object):
                The caller's value to check for a condition value within.

            condition_value (object):
                The value to check within the lookup value.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value does not contain the condition value.

        Raises:
            TypeError:
                Either the lookup or condition value was not compatible with
                the expression.
        """
        return condition_value not in match_value


class ContainsAnyOperator(BaseConditionOperator):
    """Checks if a lookup value contains any specified condition values.

    This operator checks if the provided lookup value contains any items in
    a list of condition value. It's useful for checking if a list contains
    anything from another list.

    It's equivalent to::

        if set(condition_value) & set(match_value):
            ...

    This is the opposite of :py:class:`DoesNotContainAnyOperator`.
    """

    operator_id = 'contains-any'
    name = _('Any of')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value contains any condition values.

        Args:
            match_value (object):
                The caller's value to check for condition values within.

            condition_value (object):
                The values to check within the lookup value.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value contains any of the given condition
            values.

        Raises:
            TypeError:
                Either the lookup or condition value was not compatible with
                the expression.
        """
        return bool(set(condition_value) & set(match_value))


class DoesNotContainAnyOperator(BaseConditionOperator):
    """Checks if a lookup value doesn't contain any of the specified values.

    This operator checks if the provided lookup value does not contain any
    of the provided condition values. It's useful for checking if a list
    does not contain any items from another list.

    It's equivalent to::

        if not (set(condition_value) & set(match_value)):
            ...

    This is the opposite of :py:class:`ContainsAnyOperator`.
    """

    operator_id = 'does-not-contain-any'
    name = _('Not any of')

    def matches(self, match_value, condition_value, **kwargs):
        """Return if the lookup value doesn't contain any condition values.

        Args:
            match_value (object):
                The caller's value to check for a condition value within.

            condition_value (object):
                The values to check within the lookup value.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value does not contain any of the condition
            value.

        Raises:
            TypeError:
                Either the lookup or condition value was not compatible with
                the expression.
        """
        return not bool(set(condition_value) & set(match_value))


class StartsWithOperator(BaseConditionOperator):
    """An operator that checks if a string starts with another string.

    This operator checks if the lookup value (assumed to be a string) starts
    with the condition value (also a string).

    It's equivalent to::

        if match_value.startswith(condition_value):
            ...
    """

    operator_id = 'starts-with'
    name = _('Starts with')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value starts with the condition value.

        Args:
            match_value (unicode):
                The caller's value to check.

            condition_value (unicode):
                The value to check at the start of the lookup value string.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value starts with the condition value.

        Raises:
            TypeError:
                Either the lookup or condition value was not a string (or
                string-like object).
        """
        try:
            return match_value.startswith(condition_value)
        except AttributeError:
            raise TypeError(
                _('Lookup value %r does not support startswith()'))


class EndsWithOperator(BaseConditionOperator):
    """An operator that checks if a string ends with another string.

    This operator checks if the lookup value (assumed to be a string) ends
    with the condition value (also a string).

    It's equivalent to::

        if match_value.endswith(condition_value):
            ...
    """

    operator_id = 'ends-with'
    name = _('Ends with')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value ends with the condition value.

        Args:
            match_value (unicode):
                The caller's value to check.

            condition_value (unicode):
                The value to check at the end of the lookup value string.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value ends with the condition value.

        Raises:
            TypeError:
                Either the lookup or condition value was not a string (or
                string-like object).
        """
        try:
            return match_value.endswith(condition_value)
        except AttributeError:
            raise TypeError(
                _('Lookup value %r does not support startswith()'))


class GreaterThanOperator(BaseConditionOperator):
    """An operator that checks if a number is greater than a value.

    This operator checks if the lookup value (assumed to be an integer or
    similar) is greater than the condition value.

    It's equivalent to::

        if match_value > condition_value:
            ...
    """

    operator_id = 'greater-than'
    name = _('Greater than')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value is greater than the condition value.

        Args:
            match_value (object):
                The caller's value to check.

            condition_value (object):
                The value that the lookup value must be greater than.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value is greater than the condition value.
        """
        return match_value > condition_value


class LessThanOperator(BaseConditionOperator):
    """An operator that checks if a number is less than a value.

    This operator checks if the lookup value (assumed to be an integer or
    similar) is less than the condition value.

    It's equivalent to::

        if match_value < condition_value:
            ...
    """

    operator_id = 'less-than'
    name = _('Less than')

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value is less than the condition value.

        Args:
            match_value (object):
                The caller's value to check.

            condition_value (object):
                The value that the lookup value must be less than.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value is less than the condition value.
        """
        return match_value < condition_value


class MatchesRegexOperator(BaseConditionOperator):
    """An operator that checks if a value matches against a regex.

    It's equivalent to::

        if condition_value.match(match_value):
            ...
    """

    operator_id = 'matches-regex'
    name = _('Matches regex')
    value_field = ConditionValueRegexField()

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value matches the condition's regex.

        Args:
            match_value (unicode):
                The caller's value to check.

            condition_value (re.RegexObject):
                The regex value that the lookup value must match.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value matches the regex in the condition.
        """
        return condition_value.match(match_value) is not None


class DoesNotMatchRegexOperator(BaseConditionOperator):
    """An operator that checks if a value does not match against a regex.

    It's equivalent to::

        if not condition_value.match(match_value):
            ...
    """

    operator_id = 'does-not-match-regex'
    name = _('Does not match regex')
    value_field = ConditionValueRegexField()

    def matches(self, match_value, condition_value, **kwargs):
        """Return whether the lookup value doesn't match the condition's regex.

        Args:
            match_value (unicode):
                The caller's value to check.

            condition_value (re.RegexObject):
                The regex value that the lookup value must not match.

            **kwargs (dict):
                Unused extra keyword arguments.

        Returns:
            bool:
            ``True`` if the lookup value doesn't match the regex in the
            condition.
        """
        return condition_value.match(match_value) is None


class ConditionOperators(OrderedRegistry):
    """Represents a list of operators for a condition choice.

    This stores a list of operators that can be used for condition choices.
    It can be used in one of two ways:

    1. Created dynamically, taking a list of :py:class:`BaseConditionOperator`
       subclasses as arguments.
    2. Subclassed, with :py:attr:`operator_classes` set to a list of
       :py:class:`BaseConditionOperator` subclasses.

    This works as a :py:ref:`registry <registry-guides>`, allowing additional
    choices to be added dynamically by extensions or other code.
    """

    #: A list of default operators.
    #:
    #: This is only used if a list of operators is not passed to the
    #: constructor.
    operator_classes = []

    lookup_attrs = ('operator_id',)
    lookup_error_class = ConditionOperatorNotFoundError
    already_registered_error_class = ConditionOperatorConflictError

    default_errors = dict(DEFAULT_ERRORS, **{
        ALREADY_REGISTERED: _(
            'Could not register condition operator %(item)s: This operator is '
            'already registered or its ID conflicts with another operator.'
        ),
        ATTRIBUTE_REGISTERED: _(
            'Could not register condition operator %(item)s: Another operator '
            '%(duplicate)s) is already registered with the same ID.'
        ),
        NOT_REGISTERED: _(
            'No operator was found matching "%(attr_value)s".'
        ),
        UNREGISTER: _(
            'Could not unregister condition operator %(item)s: This condition '
            'was not yet registered.'
        ),
    })

    def __init__(self, operators=[], *args, **kwargs):
        """Initialize the list of operators.

        Args:
            operators (list of type, optional):
                A list of :py:class:`BaseConditionOperator` subclasses. If this
                is provided, any value set for :py:attr:`operator_classes` will
                be ignored.
        """
        super(ConditionOperators, self).__init__(*args, **kwargs)

        self._operators = operators or self.operator_classes

    def get_operator(self, operator_id, choice):
        """Return an operator instance with the given ID.

        Args:
            operator_id (unicode):
                The ID of the operator to retrieve.

            choice (djblets.conditions.choices.BaseConditionChoice):
                The choice that will own this operator instance.

        Returns:
            BaseConditionOperator:
            The operator instance matching the ID.

        Raises:
            djblets.conditions.errors.ConditionOperatorNotFoundError:
                No operator was found that matched the given ID.
        """
        operator_cls = self.get('operator_id', operator_id)

        return operator_cls(choice)

    def get_defaults(self):
        """Return the default operators for the list.

        This is used internally by the parent registry classa, and is based on
        the list of operators provided to the constructor or the value for
        :py:attr:`operator_classes`.

        Returns:
            list of type:
            The default list of operators.
        """
        return self._operators



"""HTML-related utilities."""

from django.utils.functional import lazy
from django.utils.safestring import mark_safe


#: Lazily mark text as safe.
#:
#: This is useful if you need a lazily-translated string (such as with
#: :py:func:`~django.utils.translation.gettext_lazy`) to be marked safe.
#:
#: Args:
#:     text (str):
#:         The text to mark safe.
#:
#: Returns:
#:     django.utils.functional.Promise:
#:     A promise representing the safe text.
mark_safe_lazy = lazy(mark_safe, str)



"""Unit tests for djblets.datagrid."""

from datetime import datetime, timedelta

import kgb
from django.conf import settings
from django.contrib.auth.models import Group, User
from django.http import HttpRequest
from django.test.client import RequestFactory
from django.utils.encoding import force_str
from django.utils.safestring import SafeText

from djblets.datagrid.grids import (CheckboxColumn, Column, DataGrid,
                                    DateTimeSinceColumn, StatefulColumn,
                                    logger)
from djblets.deprecation import RemovedInDjblets40Warning
from djblets.testing.testcases import TestCase
from djblets.util.dates import get_tz_aware_utcnow


class GroupDataGrid(DataGrid):
    objid = Column('ID', link=True, sortable=True, field_name='id')
    name = Column('Group Name', link=True, sortable=True, expand=True)

    def __init__(self, request):
        super(GroupDataGrid, self).__init__(
            request=request,
            queryset=Group.objects.order_by('pk'),
            title='All Groups')
        self.default_sort = []
        self.default_columns = ['objid', 'name']


class CheckboxColumnTests(TestCase):
    """Unit tests for djblets.datagrid.grids.CheckboxColumn."""

    def test_initial_state(self):
        """Testing CheckboxColumn initial state"""
        column = CheckboxColumn(checkbox_name='my_checkbox&',
                                detailed_label='<Select Rows>')

        self.assertHTMLEqual(
            column.label,
            '<input class="datagrid-header-checkbox" '
            ' type="checkbox" data-checkbox-name="my_checkbox&amp;">')
        self.assertHTMLEqual(
            column.detailed_label_html,
            '<input type="checkbox"> &lt;Select Rows&gt;')
        self.assertEqual(column.detailed_label, '<Select Rows>')
        self.assertEqual(column.checkbox_name, 'my_checkbox&')

    def test_render_data_with_selected(self):
        """Testing CheckboxColumn.render_data with selected object"""
        group = Group.objects.create(name='test')

        column = CheckboxColumn(checkbox_name='my_checkbox&',
                                detailed_label='<Select Rows>')
        column.is_selected = lambda *args, **kwargs: True

        self.assertHTMLEqual(
            column.render_data(state=None, obj=group),
            '<input type="checkbox" data-object-id="1"'
            ' data-checkbox-name="my_checkbox&amp;" checked="true">')

    def test_render_data_with_unselected(self):
        """Testing CheckboxColumn.render_data with unselected object"""
        group = Group.objects.create(name='test')

        column = CheckboxColumn(checkbox_name='my_checkbox&',
                                detailed_label='<Select Rows>')

        self.assertHTMLEqual(
            column.render_data(state=None, obj=group),
            '<input type="checkbox" data-object-id="1"'
            ' data-checkbox-name="my_checkbox&amp;">')


class DateTimeSinceColumnTests(TestCase):
    """Unit tests for djblets.datagrid.grids.DateTimeSinceColumn."""

    def test_render_data(self):
        """Testing DateTimeSinceColumn.render_data"""
        class DummyObj:
            time = None

        column = DateTimeSinceColumn('Test', field_name='time')
        state = StatefulColumn(None, column)

        if settings.USE_TZ:
            now = get_tz_aware_utcnow()
        else:
            now = datetime.now()

        obj = DummyObj()
        obj.time = now
        self.assertEqual(column.render_data(state, obj), '0\xa0minutes ago')

        obj.time = now - timedelta(days=5)
        self.assertEqual(column.render_data(state, obj), '5\xa0days ago')

        obj.time = now - timedelta(days=7)
        self.assertEqual(column.render_data(state, obj), '1\xa0week ago')


class DataGridTests(kgb.SpyAgency, TestCase):
    """Unit tests for djblets.datagrid.grids.DataGrid."""

    def setUp(self):
        super(DataGridTests, self).setUp()

        Group.objects.bulk_create(
            Group(name='Group %02d' % i)
            for i in range(1, 100)
        )

        self.user = User(username='testuser')
        self.request = HttpRequest()
        self.request.user = self.user
        self.datagrid = GroupDataGrid(self.request)

    def test_render_listview(self):
        """Testing DataGrid.render_listview"""
        result = self.datagrid.render_listview()

        self.assertIsInstance(result, SafeText)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">', result)
        self.assertIn('<div class="datagrid-main">', result)
        self.assertInHTML('<td colspan="2">Group 01</td>', result)
        self.assertInHTML('<td colspan="2">Group 50</td>', result)
        self.assertIn('<div class="paginator">', result)
        self.assertIn('<div class="datagrid-menu', result)

    def test_render_listview_to_response(self):
        """Testing DataGrid.render_listview_to_response"""
        response = self.datagrid.render_listview_to_response()
        self.assertEqual(response.status_code, 200)

        content = force_str(response.content)

        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">',
                      content)
        self.assertIn('<div class="datagrid-main">', content)
        self.assertInHTML('<td colspan="2">Group 01</td>', content)
        self.assertInHTML('<td colspan="2">Group 50</td>', content)
        self.assertIn('<div class="paginator">', content)
        self.assertIn('<div class="datagrid-menu', content)

    def test_render_paginator(self):
        """Testing DataGrid.render_paginator"""
        self.datagrid.load_state()
        content = self.datagrid.render_paginator()

        self.assertHTMLEqual(
            content,
            '<div class="paginator">'
            ' <span class="current-page">1</span>'
            ' <a href="?page=2" title="Page 2">2</a>'
            ' <a href="?page=2" title="Next Page">&gt;</a>'
            ' <span class="page-count">2 pages&nbsp;</span>'
            '</div>')

    def test_load_state_with_sort_ascending(self):
        """Testing DataGrid.load_state with ascending sort"""
        self.request.GET['sort'] = 'name,objid'
        self.datagrid.load_state()

        self.assertEqual(self.datagrid.sort_list, ['name', 'objid'])
        self.assertEqual(len(self.datagrid.rows), self.datagrid.paginate_by)
        self.assertEqual(self.datagrid.rows[0]['object'].name, 'Group 01')
        self.assertEqual(self.datagrid.rows[1]['object'].name, 'Group 02')
        self.assertEqual(self.datagrid.rows[2]['object'].name, 'Group 03')

        # Exercise the code paths when rendering.
        result = self.datagrid.render_listview()
        self.assertIsInstance(result, SafeText)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">', result)

    def test_load_state_with_sort_descending(self):
        """Testing DataGrid.load_state with descending sort"""
        self.request.GET['sort'] = '-name'
        self.datagrid.load_state()

        self.assertEqual(self.datagrid.sort_list, ['-name'])
        self.assertEqual(len(self.datagrid.rows), self.datagrid.paginate_by)
        self.assertEqual(self.datagrid.rows[0]['object'].name, 'Group 99')
        self.assertEqual(self.datagrid.rows[1]['object'].name, 'Group 98')
        self.assertEqual(self.datagrid.rows[2]['object'].name, 'Group 97')

        # Exercise the code paths when rendering.
        result = self.datagrid.render_listview()
        self.assertIsInstance(result, SafeText)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">', result)

    def test_load_state_with_custom_column_orders(self):
        """Testing DataGrid.load_state with custom column orders"""
        self.request.GET['columns'] = 'objid'
        self.datagrid.load_state()

        self.assertEqual(len(self.datagrid.rows), self.datagrid.paginate_by)
        self.assertEqual(len(self.datagrid.rows[0]['cells']), 1)

        # Exercise the code paths when rendering.
        result = self.datagrid.render_listview()
        self.assertIsInstance(result, SafeText)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">', result)

    def test_load_state_with_load_extra_state_fields(self):
        """Testing DataGrid.load_state with load_extra_state returning fields
        """
        class MyProfile(object):
            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def get_user_profile(self):
                return my_profile

            def load_extra_state(self, profile):
                return ['field1', 'field2']

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request)
        datagrid.load_state()

        self.assertSpyCalledWith(
            my_profile.save,
            update_fields=['field1', 'field2'])

    def test_load_state_with_load_extra_state_no_fields(self):
        """Testing DataGrid.load_state with load_extra_state returning no
        fields
        """
        class MyProfile(object):
            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def get_user_profile(self):
                return my_profile

            def load_extra_state(self, profile):
                return []

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request)
        datagrid.load_state()

        self.assertSpyNotCalled(my_profile.save)

    def test_load_state_with_load_extra_state_true(self):
        """Testing DataGrid.load_state with load_extra_state returning True"""
        class MyProfile(object):
            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def get_user_profile(self):
                return my_profile

            def load_extra_state(self, profile):
                return True

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request)

        message = (
            "TestDataGrid.load_extra_state() returned a <class 'bool'>, but "
            "must return a list of field names to save (or an empty list). "
            "This will be required in Djblets 4.0."
        )

        with self.assertWarns(RemovedInDjblets40Warning, message):
            datagrid.load_state()

        self.assertSpyCalled(my_profile.save)

    def test_load_state_with_load_extra_state_false(self):
        """Testing DataGrid.load_state with load_extra_state returning False"""
        class MyProfile(object):
            def save(self, **kwargs):
                pass

        class TestDataGrid(GroupDataGrid):
            def get_user_profile(self):
                return my_profile

            def load_extra_state(self, profile):
                return False

        my_profile = MyProfile()
        self.spy_on(my_profile.save)

        datagrid = TestDataGrid(request=self.request)

        message = (
            "TestDataGrid.load_extra_state() returned a <class 'bool'>, but "
            "must return a list of field names to save (or an empty list). "
            "This will be required in Djblets 4.0."
        )

        with self.assertWarns(RemovedInDjblets40Warning, message):
            datagrid.load_state()

        self.assertSpyNotCalled(my_profile.save)

    def test_precompute_objects_with_unsortable_column_ascending(self):
        """Testing DataGrid.precompute_objects with improper sort key
        in ascending order
        """
        self.request.GET['sort'] = 'garbagetrash'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, [])

    def test_precompute_objects_with_unsortable_column_descending(self):
        """Testing DataGrid.load_state with improper sort key in
        descending order
        """
        self.request.GET['sort'] = '-garbagetrash'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, [])

    def test_precompute_objects_with_mixed_column_ascending(self):
        """Testing DataGrid.load_state with one proper and one improper sort
        key in ascending order
        """
        self.request.GET['sort'] = 'garbagetrash,name'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, ['name'])

    def test_precompute_objects_with_mixed_column_descending(self):
        """Testing DataGrid.load_state with one proper and one improper sort
        key in descending order
        """
        self.request.GET['sort'] = '-garbagetrash,-name'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, ['-name'])

    def test_post_process_queryset_with_select_related(self):
        """Testing DataGrid.post_process_queryset with chained select_related
        calls
        """
        class Column1(Column):
            def augment_queryset(self, state, queryset):
                return queryset.select_related('foo')

        class Column2(Column):
            def augment_queryset(self, state, queryset):
                return queryset.select_related('foo__bar')

        class Column3(Column):
            def augment_queryset(self, state, queryset):
                return queryset.select_related('foo', 'foo__abc')

        class TestDataGrid(DataGrid):
            col1 = Column1()
            col2 = Column2()
            col3 = Column3()

            def __init__(self, *args, **kwargs):
                super(TestDataGrid, self).__init__(*args, **kwargs)

                self.default_columns = ['col1', 'col2', 'col3']
                self.default_sort = []

            def post_process_queryset(self, queryset):
                return super(TestDataGrid, self).post_process_queryset(
                    queryset.select_related('baz'))

        grid = TestDataGrid(self.request, queryset=Group.objects.all())
        grid.columns = [
            grid.get_stateful_column(grid.get_column(name))
            for name in grid.default_columns
        ]

        queryset = grid.post_process_queryset(grid.queryset)

        self.assertEqual(
            queryset.query.select_related,
            {
                'foo': {
                    'bar': {},
                    'abc': {},
                },
                'baz': {},
            })

    def test_post_process_queryset_sandboxes_errors(self):
        """Testing DataGrid.post_process_queryset with column.augment_queryset
        raises exception
        """
        class BadColumn(Column):
            def augment_queryset(self, state, queryset):
                raise Exception

        column = BadColumn(id='test')
        self.datagrid.add_column(column)

        self.spy_on(column.augment_queryset)
        self.spy_on(logger.exception)

        try:
            self.datagrid.columns = [
                self.datagrid.get_stateful_column(column),
            ]
            self.datagrid.post_process_queryset(queryset=Group.objects.all())
        finally:
            self.datagrid.remove_column(column)

        self.assertTrue(column.augment_queryset.called)
        self.assertIn('Error when calling augment_queryset for DataGrid '
                      'Column',
                      logger.exception.last_call.args[0])

    def test_get_stateful_column_sandboxes_errors(self):
        """Testing DataGrid.get_stateful_column when column.setup_state
        raises exception
        """
        class BadColumn(Column):
            def setup_state(self, state):
                raise Exception

        column = BadColumn(id='test')
        self.datagrid.add_column(column)

        self.spy_on(column.setup_state)
        self.spy_on(logger.exception)

        try:
            self.datagrid.get_stateful_column(column=column)
        finally:
            self.datagrid.remove_column(column)

        self.assertTrue(column.setup_state.called)
        self.assertIn('Error when calling setup_state for DataGrid Column',
                      logger.exception.last_call.args[0])

    def test_get_sort_field_columns_sandboxes_errors(self):
        """Testing DataGrid.get_sort_field_columns when column.get_sort_field
        raises exception
        """
        class BadColumn(Column):
            def get_sort_field(self, state):
                raise Exception

        column = BadColumn(id='test', sortable=True)
        self.datagrid.add_column(column)

        self.datagrid.sort_list = ['test']
        self.datagrid.default_columns = ['objid', 'test']

        self.spy_on(column.get_sort_field)
        self.spy_on(logger.exception)

        try:
            self.datagrid.precompute_objects()
        finally:
            self.datagrid.remove_column(column)

        self.assertTrue(column.get_sort_field.called)
        self.assertIn('Error when calling get_sort_field for DataGrid Column',
                      logger.exception.last_call.args[0])


class ColumnTests(kgb.SpyAgency, TestCase):
    """Unit tests for djblets.datagrid.grids.Column."""

    def test_render_cell_sandboxes_errors(self):
        """Testing Column.render_cell when column.render_data raises exception
        """
        class BadColumn(Column):
            def render_data(self, state, obj):
                raise Exception

        request = RequestFactory().request()
        datagrid = DataGrid(request=request,
                            queryset=Group.objects.all())

        column = BadColumn(id='test')
        self.spy_on(column.render_data)
        self.spy_on(logger.exception)

        rendered = column.render_cell(state=StatefulColumn(datagrid=datagrid,
                                                           column=column),
                                      obj=None,
                                      render_context=None)
        self.assertHTMLEqual(rendered, '<td></td>')

        self.assertTrue(column.render_data.called)
        self.assertIn('Error when calling render_data for DataGrid Column',
                      logger.exception.last_call.args[0])
